import contextlib

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.mail import EmailMessage
from django.utils import timezone
from twilio.base.exceptions import TwilioException, TwilioRestException
from twilio.rest import Client

from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_sites.site import SiteNotRegistered, sites

from ..site_notifications import site_notifications
from ..stubs import NotificationModelStub
from ..utils import get_email_contacts, get_email_enabled


class NotificationError(Exception):
    pass


LIVE_SYSTEM: bool = getattr(settings, "LIVE_SYSTEM", False)


class Notification:
    """A generic class to generate a notification on a condition"""

    # app_name: str = None
    name: str | None = None
    display_name: str | None = None

    sms_client = Client

    email_from: str = get_email_contacts("data_manager")
    email_to: list[str] | None = None  # usually a mailing list address
    email_message_cls = EmailMessage

    email_body_template: str = (
        "\n\nDo not reply to this email\n\n"
        "{test_body_line}"
        "A report has been submitted for patient "
        "{subject_identifier} "
        "at site '{site_name}' which may require "
        "your attention.\n\n"
        "Title: {display_name}\n\n"
        "You received this message because you are subscribed to receive these "
        "notifications in your user profile.\n\n"
        "{test_body_line}"
        "Thanks."
    )
    email_subject_template: str = (
        "{test_subject_line}{protocol_name}: {display_name} for {subject_identifier}"
    )
    email_footer_template: str = (
        "\n\n-----------------\n"
        'To unsubscribe remove "{display_name}" from your chosen '
        "email notifications in your user profile.\n\n"
        "{name}\n"
        "{message_reference}\n"
        "{message_datetime} (UTC)"
    )
    email_test_body_line: str = "THIS IS A TEST MESSAGE. NO ACTION IS REQUIRED\n\n"
    email_test_subject_line: str = "TEST/UAT -- "

    sms_template: str = (
        "{test_line}{protocol_name}: Report '{display_name}' for "
        "patient {subject_identifier} "
        "at site '{site_name}' may require "
        "your attention. Login to review."
    )
    sms_test_line: str = "TEST MESSAGE. NO ACTION REQUIRED - "

    def __init__(self) -> None:
        self._notification_enabled: bool | None = None
        self._template_opts: dict = {}
        self.email_to = self.email_to or self.default_email_to
        self.test_message: bool = False
        if not LIVE_SYSTEM:
            self.email_to = [f"test.{email}" for email in self.email_to]
            self.test_message = True

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}:name='{self.name}', "
            f"display_name='{self.display_name}'>"
        )

    def __str__(self) -> str:
        return f"{self.name}: {self.display_name}"

    @property
    def default_email_to(self) -> list[str]:
        return [f"{self.name}.{settings.APP_NAME}@mg.clinicedc.org"]

    def notify(
        self,
        force_notify=None,
        use_email=None,
        use_sms=None,
        email_body_template=None,
        **kwargs,
    ) -> bool:
        """Notify / send an email and/or SMS.

        Main entry point.

        This notification class (me) knows from whom and to whom the
        notifications will be sent.

        See signals and kwargs are:
            * history_instance
            * instance
            * user
        """
        email_sent: int | None = None
        sms_sent: dict | None = None
        use_email = use_email or get_email_enabled()
        use_sms = use_sms or getattr(settings, "TWILIO_ENABLED", False)
        if force_notify or self._notify_on_condition(**kwargs):
            if use_email:
                email_body_template = (
                    email_body_template or self.email_body_template
                ) + self.email_footer_template
                email_sent = self.send_email(email_body_template=email_body_template, **kwargs)
            if use_sms:
                sms_sent = self.send_sms(**kwargs)
            self.post_notification_actions(email_sent=email_sent, sms_sent=sms_sent, **kwargs)
        return bool(email_sent or sms_sent)

    def notify_on_condition(self, **kwargs) -> bool:  # noqa: ARG002
        """Override to conditionally return True if the notification
        should be sent by email and/or sms.

        A return value of `False` means nothing will be sent.
        """
        return True

    def _notify_on_condition(self, test_message=None, **kwargs) -> bool:
        """Returns the value of `notify_on_condition` or False."""
        if test_message:
            return True
        return self.enabled and self.notify_on_condition(**kwargs)

    def post_notification_actions(self, **kwargs):
        pass

    @property
    def enabled(self) -> bool:
        """Returns True if this notification is enabled based on the value
        of Notification model instance.

        Note: Notification names/display_names are persisted in the
        "Notification" model where each mode instance can be flagged
        as enabled or not, and are selected/subscribed to by
        each user in their user profile.

        See also: `site_notifications.update_notification_list`
        """
        if not self._notification_enabled:
            self._notification_enabled = self.notification_model.enabled
        return self._notification_enabled

    @property
    def notification_model(self) -> NotificationModelStub:
        """Returns the Notification 'model' instance associated
        with this notification.
        """
        notification_model_cls = django_apps.get_model("edc_notification.notification")
        # trigger exception if this class is not registered.
        site_notifications.get(self.name)
        try:
            notification_model = notification_model_cls.objects.get(name=self.name)
        except ObjectDoesNotExist:
            site_notifications.update_notification_list()
            notification_model = notification_model_cls.objects.get(name=self.name)
        return notification_model

    def get_display_name(self, **kwargs):  # noqa: ARG002
        return self.display_name

    def get_template_options(self, instance=None, test_message=None, **kwargs):
        """Returns a dictionary of message template options.

        Extend using `extra_template_options`.
        """
        test_message = test_message or self.test_message
        template_options = dict(
            name=self.name,
            protocol_name=ResearchProtocolConfig().protocol_name,
            display_name=self.get_display_name(
                instance=instance, test_message=test_message, **kwargs
            ),
            email_from=self.email_from,
            test_subject_line=(self.email_test_subject_line if test_message else "").strip(),
            test_body_line=self.email_test_body_line if test_message else "",
            test_line=self.sms_test_line if test_message else "",
            message_datetime=timezone.now(),
            message_reference="",
        )
        if "subject_identifier" not in template_options:
            with contextlib.suppress(AttributeError):
                template_options.update(subject_identifier=instance.subject_identifier)
        if "site_name" not in template_options:
            try:
                site_name = sites.get(instance.site.id).description
            except AttributeError:
                pass
            except SiteNotRegistered as e:
                raise NotificationError(e) from e
            else:
                template_options.update(site_name=site_name.replace("_", " ").title())
        return template_options

    def send_email(
        self,
        email_to: list[str] | None = None,
        fail_silently: bool | None = None,
        email_body_template: str | None = None,
        **kwargs,
    ) -> int:
        kwargs.update(**self.get_template_options(**kwargs))
        subject = self.email_subject_template.format(**kwargs)
        body = (email_body_template or self.email_body_template).format(**kwargs)
        email = self.email_message_cls(
            subject, body, self.email_from, email_to or self.email_to
        )
        return email.send(fail_silently)

    def send_sms(
        self,
        fail_silently: bool | None = None,
        sms_recipient: str | None = None,
        **kwargs,
    ) -> dict:
        status = {}
        if self.sms_sender:
            kwargs.update(**self.get_template_options(**kwargs))
            body = self.sms_template.format(**kwargs)
            try:
                client = self.sms_client()
            except (TwilioRestException, TwilioException):
                if not fail_silently:
                    raise
            else:
                recipients = [sms_recipient] if sms_recipient else self.sms_recipients
                for recipient in recipients:
                    try:
                        message = client.messages.create(
                            from_=self.sms_sender, to=recipient, body=body
                        )
                    except (TwilioRestException, TwilioException):
                        if not fail_silently:
                            raise
                    else:
                        status.update({recipient: message.sid})
        return status

    @property
    def sms_sender(self):
        try:
            sender = settings.TWILIO_SENDER
        except AttributeError:
            sender = None
        return sender

    @property
    def sms_recipients(self):
        """Returns a list of recipients subscribed to receive SMS's
        for this "notifications" class.

        See also: edc_auth.UserProfile.
        """
        sms_recipients = []
        user_profile_model_cls = django_apps.get_model("edc_auth.UserProfile")
        for user_profile in user_profile_model_cls.objects.filter(
            user__is_active=True, user__is_staff=True
        ):
            try:
                user_profile.sms_notifications.get(name=self.name)
            except ObjectDoesNotExist:
                pass
            else:
                if user_profile.mobile:
                    sms_recipients.append(user_profile.mobile)
        return sms_recipients

    @property
    def test_template_options(self):
        return dict(subject_identifier="123456910", site_name="Gaborone")

    def send_test_email(self, email_to: str):
        """Sends a test message to "email_to".

        For example:

            from edc_notification.notification import Notification

            notification = Notification()
            notification.send_test_email('someone@example.com')
        """
        self.notify(
            force_notify=True,
            test_message=True,
            email_to=[email_to],
            **self.test_template_options,
        )

    def send_test_sms(self, sms_recipient: str):
        """Sends a test message to "email_to".

        For example:

            from edc_notification.notification import Notification

            notification = Notification()
            notification.send_test_sms(sms_recipient='+123456789')
        """
        self.notify(
            force_notify=True,
            test_message=True,
            sms_recipient=sms_recipient,
            **self.test_template_options,
        )



lab_navbar_codenames = [
    "edc_lab_dashboard.nav_lab_aliquot",
    "edc_lab_dashboard.nav_lab_manifest",
    "edc_lab_dashboard.nav_lab_pack",
    "edc_lab_dashboard.nav_lab_process",
    "edc_lab_dashboard.nav_lab_receive",
    "edc_lab_dashboard.nav_lab_requisition",
    "edc_lab_dashboard.nav_lab_section",
]

lab_navbar_tuples = []
for codename in lab_navbar_codenames:
    lab_navbar_tuples.append((codename, f"Can access {codename.split('.')[1]}"))


lab_dashboard_codenames = [
    "edc_lab_dashboard.view_lab_aliquot_listboard",
    "edc_lab_dashboard.view_lab_box_listboard",
    "edc_lab_dashboard.view_lab_manifest_listboard",
    "edc_lab_dashboard.view_lab_pack_listboard",
    "edc_lab_dashboard.view_lab_process_listboard",
    "edc_lab_dashboard.view_lab_receive_listboard",
    "edc_lab_dashboard.view_lab_requisition_listboard",
    "edc_lab_dashboard.view_lab_result_listboard",
]

lab_dashboard_tuples = []
for codename in lab_dashboard_codenames:
    lab_dashboard_tuples.append((codename, f"Can access {codename.split('.')[1]}"))



# Generated by Django 5.2.10 on 2026-01-25 04:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0127_historicalrepackrequest_unit_qty_processed_and_more"),
    ]

    operations = []



# Generated by Django 2.2.2 on 2019-09-22 02:52

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_sites", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="siteprofile",
            name="title",
            field=models.CharField(max_length=250, null=True),
        )
    ]



from edc_lab.constants import PACKED, SHIPPED, VERIFIED
from edc_listboard.filters import ListboardFilter, ListboardViewFilters


class PackListboardViewFilters(ListboardViewFilters):
    all = ListboardFilter(name="all", label="All", lookup={})

    verified = ListboardFilter(label="Verified", lookup={"status": VERIFIED})

    not_verified = ListboardFilter(
        label="Not verified", exclude_filter=True, lookup={"status": VERIFIED}
    )

    packed = ListboardFilter(label="Packed", lookup={"status": PACKED})

    not_packed = ListboardFilter(
        label="Packed", exclude_filter=True, default=True, lookup={"status": PACKED}
    )

    shipped = ListboardFilter(label="Shipped", lookup={"status": SHIPPED})

    not_shipped = ListboardFilter(
        label="Not shipped", exclude_filter=True, lookup={"status": SHIPPED}
    )



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, TypeVar

from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from edc_facility.facility import Facility
from edc_facility.utils import get_facility
from edc_visit_tracking.model_mixins import get_related_visit_model_attr

from ..utils import (
    get_appointment_type_model_cls,
    get_next_appointment,
    get_previous_appointment,
)

if TYPE_CHECKING:
    from edc_visit_tracking.model_mixins import VisitModelMixin

    from ..models import Appointment

    VisitModel = TypeVar("VisitModel", bound=VisitModelMixin)


class AppointmentMethodsModelError(Exception):
    pass


class AppointmentMethodsModelMixin(models.Model):
    """Mixin of methods for the appointment model only"""

    def get_appt_type_display(self: Appointment) -> str:
        return get_appointment_type_model_cls().objects.get(id=self.appt_type_id).display_name

    @property
    def facility(self: Appointment) -> Facility:
        """Returns the facility instance for this facility name"""
        return get_facility(name=self.facility_name)

    @property
    def visit_label(self: Appointment) -> str:
        return f"{self.visit_code}.{self.visit_code_sequence}"

    @classmethod
    def related_visit_model_attr(cls: Appointment) -> str:
        """Returns the reversed related visit attr"""
        return get_related_visit_model_attr(cls)

    @classmethod
    def related_visit_model_cls(cls: Appointment) -> type[VisitModel]:
        return getattr(cls, cls.related_visit_model_attr()).related.related_model

    @property
    def next_by_timepoint(self: Appointment) -> Appointment | None:
        """Returns the next appointment or None of all appointments
        for this subject for visit_code_sequence=0.
        """
        return (
            self.__class__.objects.filter(
                timepoint__gt=self.timepoint,
                visit_code_sequence=0,
                subject_identifier=self.subject_identifier,
            )
            .order_by("timepoint")
            .first()
        )

    @property
    def last_visit_code_sequence(self: Appointment) -> int | None:
        """Returns an integer, or None, that is the visit_code_sequence
        of the last appointment for this visit code that is not self.
        (ordered by visit_code_sequence).

        A sequence would be 1000.0, 1000.1, 1000.2, ...
        """
        obj = (
            self.__class__.objects.filter(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit_code=self.visit_code,
                visit_code_sequence__gt=self.visit_code_sequence,
            )
            .order_by("visit_code_sequence")
            .last()
        )
        if obj:
            return obj.visit_code_sequence
        return None

    @property
    def next_visit_code_sequence(self: Appointment) -> int:
        """Returns an integer that is the next visit_code_sequence
        or the last visit_code_sequence + 1 for this visit.

        A sequence would be 1000.0, 1000.1, 1000.2, ...
        """
        if self.last_visit_code_sequence:
            return self.last_visit_code_sequence + 1
        return self.visit_code_sequence + 1

    @property
    def previous_by_timepoint(self: Appointment) -> Appointment | None:
        """Returns the previous appointment or None by timepoint
        for visit_code_sequence=0.
        """
        return (
            self.__class__.objects.filter(
                timepoint__lt=self.timepoint,
                visit_code_sequence=0,
                subject_identifier=self.subject_identifier,
            )
            .order_by("timepoint")
            .last()
        )

    @property
    def first(self: Appointment) -> Appointment:
        """Returns the first appointment for this timepoint."""
        if self.visit_code_sequence == 0:
            return self
        return self.__class__.objects.get(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule_name,
            schedule_name=self.schedule_name,
            timepoint=self.timepoint,
            visit_code_sequence=0,
        )

    @property
    def previous(self: Appointment) -> Appointment | None:
        """Returns the previous appointment or None in this schedule
        for visit_code_sequence=0.
        """
        return get_previous_appointment(self, include_interim=False)

    @property
    def next(self: Appointment) -> Appointment | None:
        """Returns the next appointment or None in this schedule
        for visit_code_sequence=0.
        """
        return get_next_appointment(self, include_interim=False)

    @property
    def relative_previous(self: Appointment) -> Appointment | None:
        return get_previous_appointment(self, include_interim=True)

    @property
    def relative_next(self: Appointment) -> Appointment | None:
        return get_next_appointment(self, include_interim=True)

    @property
    def related_visit(self: Appointment) -> VisitModel | None:
        """Returns the related visit model for the current instance."""
        related_visit = None
        with contextlib.suppress(ObjectDoesNotExist):
            related_visit = getattr(self, self.related_visit_model_attr())
        return related_visit

    @property
    def relative_previous_with_related_visit(self: Appointment) -> Appointment | None:
        """Returns the first "previous" appointment with a related_visit
        instance.

        Considers interim appointments.

        Note: a NEW or SKIPPED_APPT will not have a related visit
        """
        appointment = self.relative_previous
        while appointment:
            if appointment.related_visit:
                break
            appointment = appointment.relative_previous
        return appointment

    @property
    def relative_next_with_related_visit(self: Appointment) -> Appointment | None:
        """Returns the first "next" appointment with a related_visit
        instance.

        Considers interim appointments.

        Note: a NEW or SKIPPED_APPT will not have a related visit
        """
        appointment = self.relative_next
        while appointment:
            if appointment.related_visit:
                break
            appointment = appointment.relative_next
        return appointment

    class Meta:
        abstract = True



# Generated by Django 5.0 on 2024-01-24 18:10

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_action_item",
            "0035_alter_actionitem_options_alter_actiontype_options_and_more",
        ),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.AlterField(
            model_name="actionitem",
            name="priority",
            field=models.CharField(
                blank=True,
                choices=[
                    ("high", "High Priority"),
                    ("medium", "Medium Priority"),
                    ("low", "Low Priority"),
                ],
                help_text="Leave blank to use default for this action type.",
                max_length=25,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="priority",
            field=models.CharField(
                choices=[
                    ("high", "High Priority"),
                    ("medium", "Medium Priority"),
                    ("low", "Low Priority"),
                ],
                default="high",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="priority",
            field=models.CharField(
                blank=True,
                choices=[
                    ("high", "High Priority"),
                    ("medium", "Medium Priority"),
                    ("low", "Low Priority"),
                ],
                help_text="Leave blank to use default for this action type.",
                max_length=25,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.site",
            ),
        ),
    ]



from .applicable_field_validator import ApplicableFieldValidator
from .base_form_validator import (
    APPLICABLE_ERROR,
    INVALID_ERROR,
    NOT_APPLICABLE_ERROR,
    NOT_REQUIRED_ERROR,
    REQUIRED_ERROR,
    InvalidModelFormFieldValidator,
    ModelFormFieldValidatorError,
)
from .form_validator import FormValidator
from .form_validator_mixins import FormValidatorMixin, ReportDatetimeFormValidatorMixin
from .many_to_many_field_validator import (
    M2M_INVALID_SELECTION,
    M2M_SELECTION_ONLY,
    ManyToManyFieldValidator,
)
from .other_specify_field_validator import OtherSpecifyFieldValidator
from .required_field_validator import RequiredFieldValidator
from .test_case_mixin import FormValidatorTestCaseMixin



# Generated by Django 6.0 on 2026-01-14 23:14

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0092_confirmationatlocationitem_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="confirmationatlocation",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Stock confirmation at location ",
                "verbose_name_plural": "Stock confirmations at location",
            },
        ),
        migrations.AlterModelOptions(
            name="confirmationatlocationitem",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Stock confirmation item at location",
                "verbose_name_plural": "Stock confirmation items at location",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalconfirmationatlocation",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Stock confirmation at location ",
                "verbose_name_plural": "historical Stock confirmations at location",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalconfirmationatlocationitem",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Stock confirmation item at location",
                "verbose_name_plural": "historical Stock confirmation items at location",
            },
        ),
        migrations.RenameField(
            model_name="confirmationatlocationitem",
            old_name="confirmation_at_site",
            new_name="confirm_at_location",
        ),
        migrations.RenameField(
            model_name="historicalconfirmationatlocationitem",
            old_name="confirmation_at_site",
            new_name="confirm_at_location",
        ),
        migrations.RenameIndex(
            model_name="confirmationatlocation",
            new_name="edc_pharmac_modifie_1070d2_idx",
            old_name="edc_pharmac_modifie_0b23c2_idx",
        ),
        migrations.RenameIndex(
            model_name="confirmationatlocation",
            new_name="edc_pharmac_user_mo_1e1158_idx",
            old_name="edc_pharmac_user_mo_2f1be6_idx",
        ),
        migrations.RenameIndex(
            model_name="confirmationatlocationitem",
            new_name="edc_pharmac_modifie_43202e_idx",
            old_name="edc_pharmac_modifie_d7c6b9_idx",
        ),
        migrations.RenameIndex(
            model_name="confirmationatlocationitem",
            new_name="edc_pharmac_user_mo_3fe769_idx",
            old_name="edc_pharmac_user_mo_0acc89_idx",
        ),
        migrations.AddField(
            model_name="confirmationatlocationitem",
            name="code",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=15,
                verbose_name="Stock code",
            ),
        ),
        migrations.AddField(
            model_name="confirmationatlocationitem",
            name="location_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=100,
                verbose_name="Location name",
            ),
        ),
        migrations.AddField(
            model_name="confirmationatlocationitem",
            name="stock_transfer_item",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.stocktransferitem",
            ),
        ),
        migrations.AddField(
            model_name="historicalconfirmationatlocationitem",
            name="code",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=15,
                verbose_name="Stock code",
            ),
        ),
        migrations.AddField(
            model_name="historicalconfirmationatlocationitem",
            name="location_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=100,
                verbose_name="Location name",
            ),
        ),
        migrations.AddField(
            model_name="historicalconfirmationatlocationitem",
            name="stock_transfer_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.stocktransferitem",
            ),
        ),
        migrations.AddField(
            model_name="historicalstock",
            name="returned",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="historicalstocktransferitem",
            name="code",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=15,
                verbose_name="Stock code",
            ),
        ),
        migrations.AddField(
            model_name="historicalstoragebinitem",
            name="code",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=15,
                verbose_name="Stock code",
            ),
        ),
        migrations.AddField(
            model_name="stock",
            name="returned",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="stocktransferitem",
            name="code",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=15,
                verbose_name="Stock code",
            ),
        ),
        migrations.AddField(
            model_name="storagebinitem",
            name="code",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=15,
                verbose_name="Stock code",
            ),
        ),
        migrations.AlterField(
            model_name="confirmationatlocationitem",
            name="stock",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.stock",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstocktransfer",
            name="from_location",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.location",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstocktransfer",
            name="to_location",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.location",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstoragebinitem",
            name="stock",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.stock",
            ),
        ),
        migrations.AlterField(
            model_name="stocktransfer",
            name="from_location",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="from_location",
                to="edc_pharmacy.location",
            ),
        ),
        migrations.AlterField(
            model_name="stocktransfer",
            name="to_location",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="to_location",
                to="edc_pharmacy.location",
            ),
        ),
        migrations.AlterField(
            model_name="stocktransferitem",
            name="stock",
            field=models.ForeignKey(
                limit_choices_to={"allocation__isnull": False},
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.stock",
            ),
        ),
        migrations.AlterField(
            model_name="storagebinitem",
            name="stock",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.stock",
            ),
        ),
    ]



from .date_estimated import IsDateEstimatedField, IsDateEstimatedFieldNa
from .duration import DurationDHField, DurationYMDField
from .hostname_modification_field import HostnameModificationField
from .identity_type_field import IdentityTypeField
from .initials_field import InitialsField
from .other_charfield import OtherCharField
from .userfield import UserField



from django.db import models

from edc_action_item.managers import (
    ActionIdentifierModelManager,
    ActionIdentifierSiteManager,
)

from ...constants import DEATH_REPORT_TMG_SECOND_ACTION


class DeathReportTmgSecondManager(ActionIdentifierModelManager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(action_item__action_type__name=DEATH_REPORT_TMG_SECOND_ACTION)


class DeathReportTmgSecondSiteManager(ActionIdentifierSiteManager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(action_item__action_type__name=DEATH_REPORT_TMG_SECOND_ACTION)


class DeathReportTmgSecondModelMixin(models.Model):
    action_name = DEATH_REPORT_TMG_SECOND_ACTION

    objects = DeathReportTmgSecondManager()

    on_site = DeathReportTmgSecondSiteManager()

    class Meta:
        abstract = True
        verbose_name = "Death Report TMG (2nd)"
        verbose_name_plural = "Death Report TMG (2nd)"



from edc_auth.site_auths import site_auths
from edc_screening.auth_objects import SCREENING, SCREENING_SUPER

from .auth_objects import codenames


def update_site_auths() -> None:
    site_auths.update_group(*codenames, name=SCREENING, no_delete=True)
    site_auths.update_group(*codenames, name=SCREENING_SUPER)


update_site_auths()



from __future__ import annotations

from django.apps import apps as django_apps


def get_rxrefill_model_cls():
    return django_apps.get_model("edc_pharmacy.rxrefill")


def get_rx_model_cls():
    return django_apps.get_model("edc_pharmacy.rx")


__all__ = ["get_rx_model_cls", "get_rxrefill_model_cls"]



from django.contrib.sites.models import Site
from django.core.validators import RegexValidator
from django.db import models

from edc_list_data.model_mixins import ListModelMixin
from edc_model.models import BaseUuidModel, HistoricalRecords


class Manager(models.Manager):
    use_in_migrations = True


class Location(ListModelMixin, BaseUuidModel):
    display_name = models.CharField(
        verbose_name="Name",
        max_length=250,
        unique=True,
        null=True,
        blank=True,
        help_text="(suggest 40 characters max.)",
    )

    site = models.ForeignKey(Site, on_delete=models.PROTECT, null=True, blank=True)

    # manages_bulk_stock = models.BooleanField(
    #     verbose_name="Manages bulk stock",
    #     default=False,
    # )
    #
    # may_receive_returned_stock = models.BooleanField(
    #     verbose_name="May receive returned stock",
    #     default=False,
    # )
    #
    contact_name = models.CharField(max_length=150, default="", blank=True)
    contact_tel = models.CharField(
        max_length=150,
        validators=[RegexValidator("[0-9]{1,15}")],
        default="",
        blank=True,
    )
    contact_email = models.EmailField(max_length=150, default="", blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def save(self, *args, **kwargs):
        if not self.display_name:
            self.display_name = self.name.capitalize()
        super().save(*args, **kwargs)

    class Meta(ListModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Location"
        verbose_name_plural = "Locations"



from clinicedc_constants import NULL_STRING
from django.db import models
from django.db.models import UniqueConstraint

from edc_model.models import BaseUuidModel
from edc_sites.model_mixins import SiteModelMixin


class IdentifierModelManager(models.Manager):
    def get_by_natural_key(self, identifier):
        return self.get(identifier=identifier)

    @property
    def formatted_sequence(self):
        """Returns a padded sequence segment for the identifier"""
        if self.is_derived:
            return ""
        return str(self.sequence_number).rjust(self.padding, "0")

    class Meta:
        abstract = True


class IdentifierModel(SiteModelMixin, BaseUuidModel):
    identifier = models.CharField(max_length=50, unique=True)

    name = models.CharField(max_length=100)

    subject_identifier = models.CharField(max_length=50, default=NULL_STRING)

    sequence_number = models.IntegerField(default=1)

    linked_identifier = models.CharField(max_length=50, default=NULL_STRING)

    device_id = models.IntegerField()

    protocol_number = models.CharField(max_length=25, default=NULL_STRING)

    model = models.CharField(max_length=100, default=NULL_STRING)

    identifier_type = models.CharField(max_length=100, default=NULL_STRING)

    identifier_prefix = models.CharField(max_length=25, default=NULL_STRING)

    objects = IdentifierModelManager()

    def __str__(self):
        return f"{self.identifier} {self.name}"

    def natural_key(self):
        return (self.identifier,)

    class Meta(BaseUuidModel.Meta):
        app_label = "edc_identifier"
        constraints = (
            UniqueConstraint(
                fields=["name", "identifier"], name="%(app_label)s_%(class)s_name_uniq"
            ),
        )
        indexes = BaseUuidModel.Meta.indexes



# Generated by Django 5.2.10 on 2026-01-25 04:01

import django.core.validators
from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0126_alter_historicalrepackrequest_item_qty_processed_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalrepackrequest",
            name="unit_qty_processed",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0.0"),
                help_text="Automatically calculated",
                max_digits=10,
                verbose_name="Unit quantity processed",
            ),
        ),
        migrations.AddField(
            model_name="repackrequest",
            name="unit_qty_processed",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0.0"),
                help_text="Automatically calculated",
                max_digits=10,
                verbose_name="Unit quantity processed",
            ),
        ),
        migrations.AlterField(
            model_name="historicalrepackrequest",
            name="container_unit_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Leave blank for default.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="historicalrepackrequest",
            name="item_qty_processed",
            field=models.IntegerField(
                default=0, verbose_name="Number of containers processed"
            ),
        ),
        migrations.AlterField(
            model_name="historicalrepackrequest",
            name="item_qty_repack",
            field=models.IntegerField(
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="Number of containers to repack",
            ),
        ),
        migrations.AlterField(
            model_name="repackrequest",
            name="container_unit_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Leave blank for default.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="repackrequest",
            name="item_qty_processed",
            field=models.IntegerField(
                default=0, verbose_name="Number of containers processed"
            ),
        ),
        migrations.AlterField(
            model_name="repackrequest",
            name="item_qty_repack",
            field=models.IntegerField(
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="Number of containers to repack",
            ),
        ),
    ]



from uuid import uuid4

from clinicedc_constants import CLOSED, OPEN
from django.conf import settings
from django.contrib import admin, messages
from django.core.exceptions import ObjectDoesNotExist
from django.urls.base import reverse
from django.utils import timezone
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_utils.text import formatted_datetime

from ..models import QueryRule
from ..rule import update_query_rules

DATA_MANAGER_ENABLED = getattr(settings, "DATA_MANAGER_ENABLED", True)


@admin.action(description=f"Toggle Active/Inactive {QueryRule._meta.verbose_name_plural}")
def toggle_active_flag(modeladmin, request, queryset):
    for obj in queryset:
        obj.active = False if obj.active else True
        obj.save()


@admin.action(description="Toggle DM Status (OPEN/CLOSED)")
def toggle_dm_status(modeladmin, request, queryset):
    for obj in queryset:
        obj.status = OPEN if obj.status != OPEN else CLOSED
        obj.save()


@admin.action(description=f"Copy {QueryRule._meta.verbose_name}")
def copy_query_rule_action(modeladmin, request, queryset):
    if queryset.count() == 1:
        obj = queryset[0]
        options = {
            k: v for k, v in obj.__dict__.items() if not k.startswith("_") and k not in ["id"]
        }
        options["title"] = f"{options['title']} (Copy)"
        options["active"] = False
        options["reference"] = uuid4()
        try:
            queryset.model.objects.get(title=options["title"])
        except ObjectDoesNotExist:
            new_obj = queryset.model.objects.create(**options)
            for o in obj.visit_schedule.all().order_by("timepoint"):
                new_obj.visit_schedule.add(o)
            for o in obj.data_dictionaries.all().order_by("number"):
                new_obj.data_dictionaries.add(o)
            msg = f'Query rule has been copied. New title is "{options["title"]}"'
            messages.add_message(request, messages.SUCCESS, msg)
        else:
            msg = f'Query rule already exists. Got "{options["title"]}"'
            messages.add_message(request, messages.ERROR, msg)
    else:
        messages.add_message(
            request,
            messages.ERROR,
            "Selecting multiple query rules is not allowed. Select just one.",
        )


@admin.action(description=f"Run selected {QueryRule._meta.verbose_name_plural}")
def update_query_rules_action(modeladmin, request, queryset):
    if not DATA_MANAGER_ENABLED:
        msg = (
            "Data manager features are currently disabled. See settings.DATA_MANAGER_ENABLED."
        )
        messages.add_message(request, messages.ERROR, msg)
    elif queryset:
        if settings.CELERY_ENABLED:
            update_query_rules.delay(pks=[o.pk for o in queryset])
            dte = timezone.now()
            taskresult_url = reverse("admin:django_celery_results_taskresult_changelist")
            msg = format_html(
                "Updating data queries in the background. "
                "Started at {}. "
                "An updated digest will be email upon completion. "
                'You may also check in <a href="{}?"'
                'task_name=update_query_rules">task results</A>. ',
                formatted_datetime(dte),
                mark_safe(taskresult_url),  # nosec B703, B308
            )
        else:
            results = update_query_rules(pks=[o.pk for o in queryset])
            msg = format_html(
                "Done updating data queries. Created {}, resolved {}.",
                mark_safe(results.get("created")),  # nosec B703, B308
                mark_safe(results.get("resolved")),  # nosec B703, B308
            )
        messages.add_message(request, messages.SUCCESS, msg)



# Generated by Django 6.0 on 2025-03-14 12:47
from django.db import migrations
from tqdm import tqdm

from edc_pharmacy.models import StorageBinItem


def update_stored_at_site(apps, schema_editor):
    for obj in tqdm(StorageBinItem.objects.all(), total=StorageBinItem.objects.count()):
        stock = obj.stock
        stock.stored_at_site = True
        stock.save(update_fields=("stored_at_site",))


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0066_historicalstock_stored_at_site_stock_stored_at_site"),
    ]

    operations = [migrations.RunPython(update_stored_at_site)]



from decimal import Decimal

from clinicedc_constants import NEW
from django.core.validators import MinValueValidator
from django.db import models
from edc_model.models import BaseUuidModel, HistoricalRecords
from sequences import get_next_value

from ...choices import ORDER_CHOICES
from ...exceptions import InvalidContainer, OrderItemError
from .container import Container
from .order import Order
from .product import Product


class Manager(models.Manager):
    use_in_migrations = True


class OrderItem(BaseUuidModel):
    order_item_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    order = models.ForeignKey(Order, on_delete=models.PROTECT, null=True, blank=False)

    product = models.ForeignKey(Product, on_delete=models.PROTECT, null=True, blank=False)

    container = models.ForeignKey(
        Container,
        on_delete=models.PROTECT,
        limit_choices_to={"may_order_as": True},
        null=True,
        blank=False,
    )

    container_unit_qty = models.DecimalField(
        verbose_name="Container unit quantity",
        null=True,
        blank=False,
        decimal_places=2,
        max_digits=10,
        validators=[MinValueValidator(Decimal("1.0"))],
        help_text=(
            "Number of units in the container(s). "
            "Must be the same for all containers in the record."
        ),
    )

    item_qty_ordered = models.IntegerField(
        verbose_name="Number of containers ordered",
        null=True,
        blank=False,
        validators=[MinValueValidator(1)],
        help_text="Number of containers",
    )

    unit_qty_ordered = models.DecimalField(
        verbose_name="Unit quantity ordered",
        null=True,
        decimal_places=2,
        max_digits=10,
        validators=[MinValueValidator(Decimal("1.0"))],
        help_text="Updated automatically (containers * unit_quantity per container)",
    )

    unit_qty_pending = models.DecimalField(
        verbose_name="Unit quantity pending",
        null=True,
        decimal_places=2,
        max_digits=10,
        validators=[MinValueValidator(Decimal("0.0"))],
        help_text="Unit quantity ordered - Unit quantity received. Updated automatically",
    )

    unit_qty_received = models.DecimalField(
        verbose_name="Unit quantity received",
        decimal_places=2,
        max_digits=10,
        null=True,
        validators=[MinValueValidator(Decimal("0.0"))],
        help_text="Updated automatically when units are received",
    )

    status = models.CharField(
        max_length=25,
        choices=ORDER_CHOICES,
        default=NEW,
        help_text="Updates in the signal",
    )

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.order_item_identifier}:{self.product.name} | {self.container.name}"

    def save(self, *args, **kwargs):
        if not self.id:
            self.order_item_identifier = f"{get_next_value(self._meta.label_lower):06d}"
            self.unit_qty_pending = self.item_qty_ordered * self.container_unit_qty
        self.unit_qty_ordered = self.item_qty_ordered * self.container_unit_qty
        if not self.order:
            raise OrderItemError("Order may not be null.")
        if not self.product:
            raise OrderItemError("Product may not be null.")
        if not self.container:
            raise OrderItemError("Container may not be null.")
        if not self.container.may_order_as:
            raise InvalidContainer(
                "Invalid container. Container is not configured for ordering. "
                f"Got {self.container}.Perhaps catch this in the form."
            )
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Order item"
        verbose_name_plural = "Order items"



# Generated by Django 6.0 on 2025-03-14 00:45

from django.db import migrations

import edc_pharmacy.models.stock.allocation


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0064_historicalstockrequest_start_datetime_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="AllocationProxy",
            fields=[],
            options={
                "verbose_name": "Allocation",
                "verbose_name_plural": "Allocation",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_pharmacy.allocation",),
            managers=[
                ("objects", edc_pharmacy.models.stock.allocation.Manager()),
            ],
        ),
    ]



from django.apps import apps as django_apps

from edc_model.stubs import BaseUuidHistoryModelStub

from ..constants import CREATE, DELETE, UPDATE
from .notification import Notification


class ModelNotification(Notification):
    """Generate a notification based on a model condition.

    Model must use the historical_model manager.

    The default condition is notify upon model creation
    modification or deletion.
    """

    model: str | None = None  # label_lower format

    model_operations: tuple[str] = (CREATE, UPDATE, DELETE)

    create_fields: tuple[str] = ("created",)
    update_fields: tuple[str] = ("modified",)

    email_body_template: str = (
        "\n\nDo not reply to this email\n\n"
        "{test_body_line}"
        "A report has been submitted for patient "
        "{subject_identifier} "
        "at site {site_name} which may require "
        "your attention.\n\n"
        "Title: {display_name}\n\n"
        "You received this message because you are subscribed to receive these "
        "notifications in your user profile.\n\n"
        "{test_body_line}"
        "Thanks."
    )
    email_subject_template: str = (
        "{test_subject_line}{protocol_name}: {model_operation} "
        "{display_name} for {subject_identifier}"
    )
    sms_template: str = (
        "{test_line}{protocol_name}: {model_operation}Report '{display_name}' for "
        "patient {subject_identifier} "
        "at site '{site_name}' may require "
        "your attention. Login to review. (See your user profile to unsubscribe.)"
    )

    def __init__(self) -> None:
        super().__init__()
        if not self.display_name:
            self.display_name = django_apps.get_model(self.model)._meta.verbose_name.title()

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}:name='{self.name}', "
            f"display_name='{self.display_name}',"
            f"model='{self.model}'>"
        )

    def __str__(self) -> str:
        return f"{self.name}: {self.display_name} ({self.model})"

    def notify_on_condition(self, instance: BaseUuidHistoryModelStub = None, **kwargs) -> bool:
        """Returns True if the condition in one of the C(r)UD methods is met."""
        if instance._meta.label_lower == self.model:
            return (
                self.is_create(instance, **kwargs)
                or self.is_update(instance, **kwargs)
                or self.is_delete(instance, **kwargs)
            )
        return False

    def is_create(self, instance, **kwargs):
        """Returns True if we are watching for a CREATE model operation, this is a
        CREATE model operation, and the custom create condition is met.
        """
        history_type = self.get_history_type(instance)
        return (
            history_type in self.model_operations
            and history_type == CREATE
            and self.notify_on_create_condition(instance, **kwargs)
        )

    def notify_on_create_condition(self, instance, **kwargs):
        """Returns a dictionary or None of {field: value, ...}
        for field values in `create_fields`.
        """
        created_fields = {}
        if self.create_fields:
            created = {}
            for field in self.create_fields:
                value = getattr(instance, field)
                created.update({field: value})
            for field, value in created.items():
                if self.field_value_condition_on_create(field, value):
                    created_fields.update({field: value})
        return created_fields or None

    def field_value_condition_on_create(self, field, current_value) -> bool:
        """Returns True or False.

        Override for a more complex evaluation.
        """
        return True

    def is_update(self, instance, **kwargs):
        """Returns True if we are watching for a UPDATE model operation, this is a
        UPDATE model operation, and the custom create condition is met.
        """
        history_type = self.get_history_type(instance)
        return (
            history_type in self.model_operations
            and history_type == UPDATE
            and self.notify_on_update_condition(instance, **kwargs)
        )

    def notify_on_update_condition(self, instance, **kwargs):
        """Returns a dictionary or None of {field: [previous_value, current_value], ...}
        for field values in `update_fields` that have changed.
        """
        changed_fields = {}
        if self.update_fields and instance.history.all().count() > 1:
            changes = {}
            for field in self.update_fields:
                values = [getattr(obj, field) for obj in self.get_history(instance)]
                changes.update({field: values[:2]})
            for field, values in changes.items():
                if self.field_value_condition_on_update(field, values[1], values[0]):
                    changed_fields.update({field: values})
        return changed_fields or None

    def field_value_condition_on_update(self, field, previous_value, current_value) -> bool:
        """Returns True if the value has changed.

        Override for a more complex evaluation.
        """
        return previous_value != current_value

    def is_delete(self, instance, **kwargs):
        history_type = instance.history.all().order_by("-history_date")[0].history_type
        return history_type in self.model_operations and history_type == DELETE

    @staticmethod
    def get_history(instance):
        """Returns the history queryset in desc order by `history_date`"""
        return instance.history.all().order_by("-history_date")

    @staticmethod
    def get_history_type(instance):
        """Returns the `history_type` of the most recent historical instance"""
        return instance.history.all().order_by("-history_date")[0].history_type

    def get_model_operation_as_text(self, instance=None, test_message=None, **kwargs):
        opts = {
            CREATE: "",
            UPDATE: "UPDATE* ",
            DELETE: "DELETED* ",
        }
        if not instance and test_message:
            return ""
        return f"{opts.get(self.get_history_type(instance))}"

    def get_template_options(self, instance=None, test_message=None, **kwargs) -> dict:
        opts = super().get_template_options(
            instance=instance, test_message=test_message, **kwargs
        )
        opts.update(
            message_reference=getattr(
                instance,
                "id",
                "test-message-id",
            ),
            model_operation=self.get_model_operation_as_text(
                instance=instance, test_message=test_message, **kwargs
            ),
        )

        return opts



# Generated by Django 5.2.10 on 2026-01-22 18:31

import _socket
import django.db.models.deletion
import django.utils.timezone
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
import simple_history.models
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0021_alter_edcpermissions_revision_alter_note_revision"),
        ("sites", "0002_alter_domain_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="HistoricalNote",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        default="",
                        editable=False,
                        help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                        max_length=75,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        default="",
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        default="",
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        default="",
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=10,
                        verbose_name="Device created",
                    ),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=10,
                        verbose_name="Device modified",
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("subject_identifier", models.CharField(max_length=50)),
                ("report_model", models.CharField(max_length=150)),
                (
                    "report_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("note", models.TextField(blank=True, default="")),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("PENDING", "Data is pending"),
                            ("not_available", "Data not available"),
                        ],
                        default="",
                        max_length=25,
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sites.site",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Note",
                "verbose_name_plural": "historical Notes",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



import json

import mempass
from django.apps import apps as django_apps
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import FileResponse, HttpResponseRedirect
from django.utils.decorators import method_decorator
from django.utils.html import format_html
from django.utils.text import slugify
from django.utils.translation import gettext as _
from django.views.generic.base import ContextMixin, View

from ..utils import write_queryset_to_secure_pdf


@method_decorator(login_required, name="dispatch")
class PrintPdfReportView(ContextMixin, View):
    """Download as PDF from the browser using Django's FileResponse.

    See also PdfIntermediateView and PdfReportModelMixin.
    """

    session_key = "model_pks"

    def get(self, request, *args, **kwargs):
        kwargs = self.get_context_data(**kwargs)
        return self.render_to_response(**kwargs)

    def post(self, request, *args, **kwargs):
        # accepts post data from form on intermediate page
        kwargs.update(phrase=request.POST.get("phrase"))
        return self.get(request, *args, **kwargs)

    def render_to_response(self, **kwargs) -> FileResponse | HttpResponseRedirect:
        """Render PDF buffer to FileResponse."""
        try:
            model_pks = json.loads(self.request.session.pop(self.session_key))
        except KeyError as e:
            # TODO: is this needed?
            messages.error(
                self.request,
                format_html(
                    _(
                        f"PDF report was not created because of an error. Got {e}. "
                        "Please try again."
                    ),
                    fail_silently=True,
                ),
            )
        else:
            password = kwargs.get("phrase") or slugify(mempass.mkpassword(2))
            app_label, model_name = kwargs.get("app_label"), kwargs.get("model_name")
            qs = django_apps.get_model(app_label, model_name).objects.filter(pk__in=model_pks)
            buffer = write_queryset_to_secure_pdf(
                queryset=qs, password=password, request=self.request
            )
            report_filename = self.get_report_filename(model_pks, app_label, model_name)
            self.message_user(report_filename=report_filename, password=password)
            return FileResponse(buffer, as_attachment=True, filename=report_filename)
        return HttpResponseRedirect("/")

    def get_report_filename(
        self, model_pks: list[str], app_label: str, model_name: str
    ) -> str:
        pdf_report_cls = django_apps.get_model(app_label, model_name).pdf_report_cls
        if len(model_pks) == 1:
            model_obj = django_apps.get_model(app_label, model_name).objects.get(
                pk=model_pks[0]
            )
            report_filename = model_obj.get_pdf_report(self.request).report_filename
        else:
            report_filename = pdf_report_cls.get_generic_report_filename()
        if not report_filename:
            raise ValueError("Cannot create file without a filename. Got report_filename=None")
        return report_filename

    def message_user(self, report_filename=None, password=None) -> None:
        messages.success(
            self.request,
            format_html(
                _(
                    "The report has been exported as a secure PDF. See downloads in "
                    "your browser. %(br)sFile: %(report_filename)s %(br)s"
                    "Pass-phrase: %(password)s"
                )
                % dict(report_filename=report_filename, password=password, br="<BR>"),
                fail_silently=True,
            ),
        )

    def head(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def options(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)



from __future__ import annotations

import contextlib
from datetime import datetime
from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta
from django.db import transaction

from ..constants import INCOMPLETE_APPT, NEW_APPT
from .unscheduled_appointment_creator import (
    CreateAppointmentError,
    UnscheduledAppointmentCreator,
)

if TYPE_CHECKING:
    from ..models import Appointment


def create_next_appointment_as_interim(
    appointment: Appointment, next_appt_datetime: datetime | None = None
) -> Appointment | None:
    return create_unscheduled_appointment(appointment, next_appt_datetime=next_appt_datetime)


def create_unscheduled_appointment(
    appointment: Appointment,
    next_appt_datetime: datetime | None = None,
) -> Appointment | None:
    """Create and return an unscheduled appointment if
    `next_appt_datetime` is within the window period of
    `appointment`.

    """
    unscheduled_appointment = None
    next_appt_datetime = next_appt_datetime or appointment.appt_datetime + relativedelta(
        days=1
    )
    # next_appointment = get_appointment_by_datetime(
    #     next_appt_datetime,
    #     appointment.subject_identifier,
    #     appointment.visit_schedule_name,
    #     appointment.schedule_name,
    #     raise_if_in_gap=False,
    # )
    # if appointment == next_appointment:
    if (
        appointment.relative_next
        and appointment.relative_next.appt_status == NEW_APPT
        and appointment.relative_next.visit_code_sequence > 0
    ):
        appointment.relative_next.delete()
    appt_status = appointment.appt_status
    appointment.appt_status = INCOMPLETE_APPT
    appointment.save_base(update_fields=["appt_status"])
    with transaction.atomic(), contextlib.suppress(CreateAppointmentError):
        unscheduled_appointment = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            facility=appointment.facility,
            suggested_appt_datetime=next_appt_datetime,
            suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
        )
    appointment.appt_status = appt_status
    appointment.save_base(update_fields=["appt_status"])
    return getattr(unscheduled_appointment, "appointment", None)



from __future__ import annotations

import copy
import sys
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.utils.module_loading import import_module, module_has_submodule

from edc_reportable.utils import reference_range_colllection_model_cls

if TYPE_CHECKING:
    from edc_lab.models import Panel


class AlreadyRegistered(Exception):
    pass


class RegistryNotLoaded(Exception):
    pass


class SiteLabsRequisitionModelError(Exception):
    pass


class SiteLabsLabProfileError(Exception):
    pass


class SiteLabs:
    panel_model = "edc_lab.panel"

    def __init__(self):
        self._registry = {}
        self.aliquot_types = {}
        self.loaded = False
        self.migrated = False
        self.requisition_models = {}

    def __repr__(self):
        return f"{self.__class__.__name__}(loaded={self.loaded})"

    def initialize(self):
        self._registry = {}
        self.aliquot_types = {}
        self.loaded = False
        self.requisition_models = {}

    @property
    def registry(self):
        if not self.loaded:
            raise RegistryNotLoaded(
                "Registry not loaded. Is AppConfig for 'edc_lab' declared in settings?."
            )
        return self._registry

    def get(self, lab_profile_name):
        """Returns a LabProfile instance."""
        if not self.loaded:
            raise RegistryNotLoaded(self)
        return self._registry.get(lab_profile_name)

    def register(self, lab_profile=None):
        """Registers a LabProfile instance using the label_lower (model)
        as the key.

            lab_profile: instance of LabProfile
        """
        if lab_profile:
            self.loaded = True
            if lab_profile.name not in self.registry:
                self.registry.update({lab_profile.name: lab_profile})
                self.registry.update({lab_profile.requisition_model: lab_profile})
                self.aliquot_types.update(**lab_profile.aliquot_types)
                self.requisition_models.update(
                    {lab_profile.requisition_model: lab_profile.requisition_model}
                )
                if self.migrated:
                    panel_model_cls = django_apps.get_model(self.panel_model)
                    self.update_panel_model(panel_model_cls=panel_model_cls)
                    self.verify_reference_range_collections(lab_profile)
            else:
                raise AlreadyRegistered(f"Lab profile {lab_profile} is already registered.")

    @property
    def lab_profiles(self):
        return {lab_profile.name: lab_profile for lab_profile in self._registry.values()}

    @property
    def panel_model_cls(self):
        return django_apps.get_model(self.panel_model)

    @property
    def panel_names(self):
        return {
            obj.name: obj.display_name
            for obj in self.panel_model_cls.objects.all().order_by("lab_profile_name")
        }

    def check_lab_profile_name_on_panels(self, panel_model_cls=None):
        """Checks if for panels referencing non-existent lab_profiles"""
        panel_model_cls = panel_model_cls or self.panel_model_cls
        for panel in panel_model_cls.objects.all().order_by("lab_profile_name"):
            if panel.lab_profile_name not in self.lab_profiles:
                raise SiteLabsLabProfileError(
                    f"Existing panel refers to an unknown lab_profile name. "
                    f"Got '{panel.lab_profile_name}'. "
                    f"Resolve the 'lab_profile_name' on model Panel "
                    f"before attempting to creating new Panel model instances."
                )

    def update_panel_model(self, panel_model_cls: type[Panel] = None) -> None:
        """Updates or creates panel mode instances.

        Initially called in the post_migrate signal.
        """
        panel_model_cls = panel_model_cls or self.panel_model_cls

        self.check_lab_profile_name_on_panels(panel_model_cls=panel_model_cls)

        for panel in panel_model_cls.objects.all().order_by("lab_profile_name"):
            if panel.lab_profile_name not in self.lab_profiles:
                raise SiteLabsLabProfileError(
                    f"Existing panel refers to an unknown lab_profile name. "
                    f"Got '{panel.lab_profile_name}'. "
                    f"Resolve the 'lab_profile_name' on model Panel "
                    f"before attempting to creating new Panel model instances."
                )

        for lab_profile in self.registry.values():
            for panel in lab_profile.panels.values():
                try:
                    panel_model_obj = panel_model_cls.objects.get(
                        name=panel.name, lab_profile_name=lab_profile.name
                    )
                except ObjectDoesNotExist:
                    panel_model_cls.objects.create(
                        name=panel.name,
                        display_name=panel.verbose_name,
                        lab_profile_name=lab_profile.name,
                    )
                else:
                    panel_model_obj.display_name = panel.verbose_name
                    panel_model_obj.save()

    def to_csv(self):
        pass

    @staticmethod
    def verify_reference_range_collections(lab_profile):
        for panel in lab_profile.panels.values():
            if (
                panel.reference_range_collection_name
                and not reference_range_colllection_model_cls()
                .objects.filter(name=panel.reference_range_collection_name)
                .exists()
            ):
                names = [
                    obj.name for obj in reference_range_colllection_model_cls().objects.all()
                ]
                raise SiteLabsLabProfileError(
                    "Unknown reference range collection. "
                    f"Collection referenced by panel `{panel}`. "
                    f"Expected one of {names}. "
                    "Got panel.reference_range_collection_name="
                    f"`{panel.reference_range_collection_name}` "
                    "Hint: Check post-migrate signal in edc_reportable."
                )

    def autodiscover(self, module_name=None, verbose=False):
        """Autodiscovers classes in the labs.py file of any
        INSTALLED_APP.
        """
        module_name = module_name or "labs"
        sys.stdout.write(f" * checking for {module_name} ...\n")
        for app in django_apps.app_configs:
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(site_labs._registry)
                    import_module(f"{app}.{module_name}")
                    sys.stdout.write(f"   - registered labs from application '{app}'\n")
                except Exception as e:
                    if f"No module named '{app}.{module_name}'" not in str(e):
                        site_labs._registry = before_import_registry
                        if module_has_submodule(mod, module_name):
                            raise
            except ModuleNotFoundError:
                pass


site_labs = SiteLabs()



# Generated by Django 2.2.11 on 2020-03-22 18:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_sites", "0002_auto_20190922_0452"),
    ]

    operations = [
        migrations.AddField(
            model_name="siteprofile",
            name="country",
            field=models.CharField(max_length=250, null=True),
        ),
    ]



from __future__ import annotations

import copy
import sys
from json.decoder import JSONDecodeError
from typing import TYPE_CHECKING, Any

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.color import color_style
from django.utils.module_loading import import_module, module_has_submodule
from requests.exceptions import ConnectionError

from .mailing_list_manager import MailingListManager
from .utils import get_email_enabled

if TYPE_CHECKING:
    from edc_action_item.action_item_notification import ActionItemNotification
    from edc_action_item.models import ActionItem

    from .notification import Notification


style = color_style()


class AlreadyRegistered(Exception):
    pass


class RegistryNotLoaded(Exception):
    pass


class NotificationNotRegistered(Exception):
    pass


class SiteNotifications:
    def __init__(self):
        self._registry: dict[str, type[ActionItemNotification | Notification]] = {}
        self.loaded = False
        self.models = {}

    def __repr__(self):
        return f"{self.__class__.__name__}(loaded={self.loaded})"

    @property
    def registry(self):
        if not self.loaded:
            raise RegistryNotLoaded(
                "Registry not loaded. Is AppConfig for 'edc_notification' "
                "declared in settings?."
            )
        return self._registry

    def get(self, name) -> type[Notification]:
        """Returns a Notification by name."""
        if not self.loaded:
            raise RegistryNotLoaded(self)
        if not self._registry.get(name):
            raise NotificationNotRegistered(f"Notification not registered. Got '{name}'.")
        return self._registry.get(name)

    def register(self, notification_cls: type[Notification] = None):
        """Registers a Notification class unique by name."""
        self.loaded = True
        display_names = [n.display_name for n in self.registry.values()]
        if (
            notification_cls.name not in self.registry
            and notification_cls.display_name not in display_names
        ):
            self.registry.update({notification_cls.name: notification_cls})

            models = getattr(notification_cls, "models", [])
            if not models and getattr(notification_cls, "model", None):
                models = [notification_cls.model]
            for model in models:
                try:
                    if notification_cls.name not in [n.name for n in self.models[model]]:
                        self.models[model].append(notification_cls)
                except KeyError:
                    self.models.update({model: [notification_cls]})
        else:
            raise AlreadyRegistered(
                f"Notification {notification_cls.name}: "
                f"{notification_cls.display_name} is already registered."
            )

    def notify(self, instance: ActionItem = None, **kwargs):
        """A wrapper to call notification.notify for each notification
        class associated with the given model instance.

        Returns a dictionary of {notification.name: model, ...}
        including only notifications sent.
        """
        notified: dict[str, Any] = {}
        for notification_cls in self.registry.values():
            notification = notification_cls()
            if notification.notify(instance=instance, **kwargs):
                notified.update({notification_cls.name: instance._meta.label_lower})
        return notified

    def update_notification_list(
        self, apps: django_apps = None, schema_editor=None, verbose=False
    ):
        """Updates the notification model to ensure all registered
        notifications classes are listed.

        Typically called from a post_migrate signal.

        Also, in tests you can register a notification and the Notification
        class (not model) will automatically call this method if the
        named notification does not exist. See notification.notify()
        """
        notification_model_cls = (apps or django_apps).get_model(
            "edc_notification.notification"
        )

        # flag all notifications as disabled and re-enable as required
        notification_model_cls.objects.all().update(enabled=False)
        if site_notifications.loaded:
            if verbose:
                sys.stdout.write(style.MIGRATE_HEADING("Populating Notification model:\n"))
            self.delete_unregistered_notifications(apps=apps)
            for name, notification_cls in site_notifications.registry.items():
                if verbose:
                    sys.stdout.write(
                        f"  * Adding '{name}': '{notification_cls().display_name}'\n"
                    )
                try:
                    obj = notification_model_cls.objects.get(name=name)
                except ObjectDoesNotExist:
                    notification_model_cls.objects.create(
                        name=name,
                        display_name=notification_cls().display_name,
                        mailing_list_address=notification_cls().email_to[0],
                        enabled=True,
                    )
                else:
                    obj.display_name = notification_cls().display_name
                    obj.mailing_list_address = notification_cls().email_to[0]
                    obj.enabled = True
                    obj.save()

    @staticmethod
    def delete_unregistered_notifications(apps: django_apps = None):
        """Delete orphaned notification model instances."""
        notification_model_cls = (apps or django_apps).get_model(
            "edc_notification.notification"
        )
        return notification_model_cls.objects.exclude(
            name__in=[n.name for n in site_notifications.registry.values()]
        ).delete()

    def create_mailing_lists(self, verbose=True):
        """Creates the mailing list for each registered notification."""
        responses = {}
        if (
            get_email_enabled()
            and self.loaded
            and settings.EMAIL_BACKEND != "django.core.mail.backends.locmem.EmailBackend"
        ):
            sys.stdout.write(style.MIGRATE_HEADING("Creating mailing lists:\n"))
            for name, notification_cls in self.registry.items():
                notification = notification_cls()
                manager = MailingListManager(
                    address=notification.email_to,
                    name=notification.name,
                    display_name=notification.display_name,
                )
                try:
                    response = manager.create()
                except ConnectionError as e:
                    sys.stdout.write(
                        style.ERROR(f"  * Failed to create mailing list {name}. Got {e}\n")
                    )
                else:
                    if verbose:
                        try:
                            message = response.json().get("message")
                        except JSONDecodeError:
                            message = response.text
                        sys.stdout.write(
                            f"  * Creating mailing list {name}. "
                            f'Got {response.status_code}: "{message}"\n'
                        )
                    responses.update({name: response})
        return responses

    @staticmethod
    def autodiscover(module_name: str = None, verbose: bool | None = False):
        """Autodiscovers classes in the notifications.py file of any
        INSTALLED_APP.
        """
        module_name = module_name or "notifications"
        verbose = True if verbose is None else verbose
        sys.stdout.write(f" * checking for {module_name} ...\n")
        for app in django_apps.app_configs:
            before_import_registry = None
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(site_notifications._registry)
                    import_module(f"{app}.{module_name}")
                    if verbose:
                        sys.stdout.write(
                            f"   - registered notifications from application '{app}'\n"
                        )
                except Exception as e:
                    if f"No module named '{app}.{module_name}'" not in str(e):
                        site_notifications._registry = before_import_registry
                        if module_has_submodule(mod, module_name):
                            raise
            except ModuleNotFoundError:
                pass


site_notifications = SiteNotifications()



from django.urls.conf import path
from django.views.generic.base import RedirectView

app_name = "edc_randomization"


urlpatterns = [
    path("", RedirectView.as_view(url="/edc_randomization/admin/"), name="home_url"),
]



# Generated by Django 5.2.10 on 2026-01-24 15:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0110_remove_container_qty_suggested_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="container",
            name="qty_default",
            field=models.DecimalField(
                decimal_places=2,
                help_text="May be adjusted at receiving. Default value for entry form",
                max_digits=10,
                null=True,
                verbose_name="Default qty",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="qty_places",
            field=models.IntegerField(
                default=0, help_text="May NOT be adjusted at receiving"
            ),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="qty_default",
            field=models.DecimalField(
                decimal_places=2,
                help_text="May be adjusted at receiving. Default value for entry form",
                max_digits=10,
                null=True,
                verbose_name="Default qty",
            ),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="qty_places",
            field=models.IntegerField(
                default=0, help_text="May NOT be adjusted at receiving"
            ),
        ),
    ]



from .consent_model_admin_mixin import ConsentModelAdminMixin, ModelAdminConsentMixin
from .requires_consent_model_admin_mixin import RequiresConsentModelAdminMixin

__all__ = [
    "ConsentModelAdminMixin",
    "ModelAdminConsentMixin",
    "RequiresConsentModelAdminMixin",
]



from __future__ import annotations

from decimal import Decimal
from typing import TYPE_CHECKING

from django import forms
from edc_form_validators import FormValidator, FormValidatorMixin

from ...models import RepackRequest

if TYPE_CHECKING:
    from ...models import Container, Stock


class RepackRequestFormValidator(FormValidator):
    def clean(self):
        pass


class RepackRequestForm(FormValidatorMixin, forms.ModelForm):
    form_validator_cls = RepackRequestFormValidator

    @property
    def from_stock(self) -> Stock | None:
        return self.cleaned_data.get("from_stock") or getattr(
            self.instance, "from_stock", None
        )

    @property
    def container(self) -> Container | None:
        return self.cleaned_data.get("container") or getattr(self.instance, "container", None)

    @property
    def container_unit_qty(self) -> Decimal | None:
        return self.cleaned_data.get("container_unit_qty") or self.container.unit_qty_default

    @property
    def override_container_unit_qty(self) -> Decimal | None:
        return (
            self.cleaned_data.get("override_container_unit_qty")
            or self.instance.override_container_unit_qty
        )

    def clean(self):
        cleaned_data = super().clean()
        if (
            self.instance.id
            and cleaned_data.get("item_qty_repack") != self.instance.item_qty_repack
        ):
            raise forms.ValidationError(
                "Number of containers to repack may not be changed after processing"
            )
        if (
            self.from_stock
            and self.container
            and cleaned_data.get("item_qty_repack") is not None
        ):
            if self.from_stock and not getattr(self.from_stock, "confirmation", None):
                raise forms.ValidationError(
                    {
                        "from_stock": (
                            "Unconfirmed stock item. Only confirmed "
                            "stock items may be used to repack"
                        )
                    }
                )
            if self.container and self.container == self.from_stock.container:
                raise forms.ValidationError(
                    {"container": "Stock is already packed in this container."}
                )

            if (
                not self.override_container_unit_qty
                and self.container_unit_qty != self.container.unit_qty_default
            ):
                raise forms.ValidationError(
                    {
                        "container_unit_qty": (
                            "Invalid. Expected default container value of "
                            f"{self.container.unit_qty_default}."
                        )
                    }
                )
            if (
                self.container_unit_qty
                and self.container_unit_qty > self.container.unit_qty_max
            ):
                raise forms.ValidationError(
                    {"container_unit_qty": "Cannot exceed container maximum unit quantity."}
                )

            if (
                self.container_unit_qty
                and self.container_unit_qty > self.from_stock.container_unit_qty
            ):
                raise forms.ValidationError(
                    {"container": "Cannot pack into larger container."}
                )
            if (
                cleaned_data.get("item_qty_repack")
                and self.instance.item_qty_processed
                and cleaned_data.get("item_qty_repack") < self.instance.item_qty_processed
            ):
                raise forms.ValidationError(
                    {
                        "item_qty_repack": "Cannot be less than the number of containers processed"
                    }
                )

            if not self.instance.id and (
                cleaned_data.get("item_qty_repack") * self.container_unit_qty
                > self.from_stock.unit_qty
            ):
                needed_qty = cleaned_data.get("item_qty_repack") * self.container_unit_qty
                on_hand_qty = self.from_stock.unit_qty
                raise forms.ValidationError(
                    {
                        "item_qty_repack": (
                            "Insufficient unit quantity to repack from this stock item. "
                            f"Need {needed_qty} units but have only {on_hand_qty} units on hand"
                        )
                    }
                )

        return cleaned_data

    class Meta:
        model = RepackRequest
        fields = "__all__"
        help_text = {  # noqa: RUF012
            "repack_identifier": "(read-only)",
        }
        widgets = {  # noqa: RUF012
            "repack_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from __future__ import annotations

import sys
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING

from django.utils import timezone

from edc_metadata import KEYED, REQUIRED
from edc_metadata.metadata_inspector import MetaDataInspector

from ..models import QueryVisitSchedule
from .query_rule_wrapper import QueryRuleWrapper

if TYPE_CHECKING:
    from ..models import QueryRule


class RuleRunner:
    def __init__(
        self,
        query_rule_obj: QueryRule = None,
        now: datetime | None = None,
        verbose: bool | None = None,
    ):
        self.query_rule_obj = query_rule_obj  # query rule model instance
        self.now = now or timezone.now()
        self.verbose = verbose

    @property
    def reference(self):
        return self.query_rule_obj.reference

    def run(self, query_rules_data: dict[str, list[QueryRuleWrapper]] | None = None):
        """Returns number of created and resolved queries after
        running all wrapped query rule objs included in
        query_rules_data.
        """
        created_counter = 0
        resolved_counter = 0
        query_rules_data = query_rules_data or self.query_rules_data
        for _, wrapped_query_rules in query_rules_data.items():
            for wrapped_query_rule in wrapped_query_rules:
                if self.verbose:
                    sys.stdout.write(f"     - running {wrapped_query_rule}\n")
                created, resolved = wrapped_query_rule.run_handler()
                if self.verbose:
                    sys.stdout.write(
                        f"       (queries created: {created}, queries resolved: {resolved})\n"
                    )
                created_counter += created
                resolved_counter += resolved
        return created_counter, resolved_counter

    def run_one(
        self,
        subject_identifier: str,
        visit_schedule_name: str,
        schedule_name: str,
        visit_code: str,
        timepoint: Decimal,
    ):
        visit_schedule_obj = QueryVisitSchedule.objects.get(
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            visit_code=visit_code,
        )
        return self.run(
            query_rules_data={
                visit_schedule_obj.visit_code: [
                    QueryRuleWrapper(
                        query_rule_obj=self.query_rule_obj,
                        subject_identifiers=[subject_identifier],
                        visit_schedule_obj=visit_schedule_obj,
                        timepoint=timepoint,
                        now=self.now,
                    )
                ]
            }
        )

    @property
    def query_rules_data(self) -> dict[str, list[QueryRuleWrapper]]:
        """Returns a dictionary of QueryRuleWrappers format
        {visit_code: [wrappper, ...]}.

        Important: Metadata must be up-to-date
        """
        query_rules = {}
        for visit_schedule_obj in self.query_rule_obj.visit_schedule.all():
            wrapped_query_rules = []
            if self.verbose:
                sys.stdout.write(
                    f"     - query {visit_schedule_obj.visit_schedule_name}."
                    f"{visit_schedule_obj.schedule_name}.{visit_schedule_obj.visit_code}"
                    f"(timepoint {visit_schedule_obj.timepoint})\n"
                )
            metadata_inspector = MetaDataInspector(
                model_cls=self.query_rule_obj.model_cls,
                visit_schedule_name=visit_schedule_obj.visit_schedule_name,
                schedule_name=visit_schedule_obj.schedule_name,
                visit_code=visit_schedule_obj.visit_code,
                timepoint=visit_schedule_obj.timepoint,
            )
            if metadata_inspector.required:
                wrapped_query_rules.append(
                    QueryRuleWrapper(
                        query_rule_obj=self.query_rule_obj,
                        subject_identifiers=metadata_inspector.required,
                        visit_schedule_obj=visit_schedule_obj,
                        timepoint=visit_schedule_obj.timepoint,
                        now=self.now,
                        entry_status=REQUIRED,
                        verbose=self.verbose,
                    )
                )
            if metadata_inspector.keyed:
                wrapped_query_rules.append(
                    QueryRuleWrapper(
                        query_rule_obj=self.query_rule_obj,
                        subject_identifiers=metadata_inspector.keyed,
                        visit_schedule_obj=visit_schedule_obj,
                        timepoint=visit_schedule_obj.timepoint,
                        now=self.now,
                        entry_status=KEYED,
                        verbose=self.verbose,
                    )
                )
            if wrapped_query_rules:
                query_rules.update({visit_schedule_obj.visit_code: wrapped_query_rules})
        return query_rules



from django.contrib.sites.models import Site
from django.db import models
from django.utils import timezone

from ..model_mixins import qa_reports_permissions


class QaReportLog(models.Model):
    username = models.CharField(max_length=100)
    site = models.ForeignKey(Site, on_delete=models.CASCADE)
    report_model = models.CharField(max_length=100)
    accessed = models.DateTimeField(default=timezone.now)

    class Meta:
        verbose_name = "QA Report Log"
        verbose_name_plural = "QA Report Logs"
        indexes = (models.Index(fields=["accessed"]),)
        default_permissions = qa_reports_permissions



from __future__ import annotations

from decimal import Decimal
from typing import TYPE_CHECKING

from celery.states import PENDING
from clinicedc_constants import CANCEL, COMPLETE, NEW, NULL_STRING
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import ProtectedError, Sum
from django.db.models.signals import post_delete, post_save, pre_delete
from django.dispatch import receiver
from django.utils.translation import gettext as _

from edc_utils.celery import get_task_result, run_task_sync_or_async

from ..constants import PARTIAL
from ..exceptions import InsufficientStockError
from ..model_mixins import StudyMedicationCrfModelMixin
from ..utils import (
    create_new_stock_on_receive,
    process_repack_request,
    update_previous_refill_end_datetime,
)
from .stock import (
    Allocation,
    Confirmation,
    ConfirmationAtLocationItem,
    DispenseItem,
    OrderItem,
    Receive,
    ReceiveItem,
    RepackRequest,
    Stock,
    StockAdjustment,
    StockRequest,
    StockRequestItem,
    StockTransferItem,
    StorageBinItem,
)

if TYPE_CHECKING:
    from .stock import Order


def update_order_status(order: Order):
    unit_qty_received = OrderItem.objects.filter(order=order).aggregate(
        unit_qty=Sum("unit_qty_received")
    )["unit_qty"] or Decimal("0.0")
    unit_qty_ordered = OrderItem.objects.filter(order=order).aggregate(
        unit_qty=Sum("unit_qty_ordered")
    )["unit_qty"] or Decimal("0.0")
    if unit_qty_received is None or unit_qty_received == Decimal("0.0"):
        order.status = NEW
    elif unit_qty_received == unit_qty_ordered:
        order.status = COMPLETE
    elif unit_qty_received > Decimal("0.0"):
        order.status = PARTIAL
    order.save(update_fields=["status"])
    order.refresh_from_db()


def update_orderitem_status(order_item: OrderItem):
    if order_item.unit_qty_received == Decimal("0.0"):
        order_item.status = NEW
    elif order_item.unit_qty_received == order_item.unit_qty_ordered:
        order_item.status = COMPLETE
    else:
        order_item.status = PARTIAL
    order_item.save(update_fields=["status"])
    order_item.refresh_from_db()


@receiver(post_save, sender=Confirmation, dispatch_uid="confirmation_on_post_save")
def confirmation_on_post_save(sender, instance, raw, created, update_fields, **kwargs):
    """Update confirmed"""
    if not raw and not update_fields:
        instance.stock.confirmed = True
        instance.stock.save(update_fields=["confirmed"])


@receiver(post_save, sender=Stock, dispatch_uid="stock_on_post_save")
def stock_on_post_save(sender, instance, raw, created, update_fields, **kwargs):
    """Update unit qty and other columns"""
    pass


@receiver(
    post_save,
    sender=StockAdjustment,
    dispatch_uid="stock_adjustment_on_post_save",
)
def stock_adjustment_on_post_save(sender, instance, raw, created, update_fields, **kwargs):
    """Update unit qty"""
    if not raw and not update_fields:
        instance.stock.unit_qty_in = instance.unit_qty_in_new
        if instance.stock.unit_qty_out > instance.stock.unit_qty_in:
            raise InsufficientStockError(
                "Invalid adjustment. Expected a value greater than or equal to "
                f"{instance.stock.unit_qty_out}. See {instance}."
            )
        instance.stock.save(update_fields=["unit_qty_in"])


@receiver(post_save, sender=OrderItem, dispatch_uid="order_item_on_post_save")
def order_item_on_post_save(sender, instance, raw, created, update_fields, **kwargs):
    if not raw and not update_fields:
        # recalculate unit_qty_pending
        instance.unit_qty_ordered = instance.item_qty_ordered * instance.container_unit_qty
        instance.unit_qty_pending = instance.unit_qty_ordered - (
            instance.unit_qty_received or Decimal("0.0")
        )
        instance.save(update_fields=["unit_qty_pending", "unit_qty_ordered"])
        instance.refresh_from_db()
        update_orderitem_status(order_item=instance)
        update_order_status(order=instance.order)


@receiver(post_save, sender=Receive, dispatch_uid="receive_on_post_save")
def receive_on_post_save(sender, instance, raw, created, update_fields, **kwargs) -> None:
    if not raw and not update_fields:
        pass


@receiver(post_save, sender=ReceiveItem, dispatch_uid="receive_item_on_post_save")
def receive_item_on_post_save(sender, instance, raw, created, update_fields, **kwargs):
    if not raw and update_fields in (["added_to_stock"], ["container_unit_qty"]):
        # update receive item_count after receive_item
        receive_items = ReceiveItem.objects.filter(receive=instance.receive)
        instance.receive.item_count = receive_items.count()
        instance.receive.save(update_fields=["item_count"])

        # update order_item after receive_item
        instance.order_item.unit_qty_received = (
            instance.order_item.receiveitem_set.all().aggregate(
                unit_qty=Sum("unit_qty_received")
            )["unit_qty"]
        ) or Decimal("0.0")

        instance.order_item.save(update_fields=["unit_qty_received"])

        update_orderitem_status(order_item=instance.order_item)

        # update order status
        update_order_status(order=instance.order_item.order)

        # add to stock
        create_new_stock_on_receive(receive_item_pk=instance.id)


@receiver(post_save, sender=StockRequest, dispatch_uid="stock_request_on_post_save")
def stock_request_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
) -> None:
    if not raw and not update_fields:
        if instance.cancel == CANCEL:
            if not Allocation.objects.filter(
                stock_request_item__stock_request=instance
            ).exists():
                instance.stockrequestitem_set.all().delete()
            else:
                instance.cancel = ""
                instance.save(update_fields=["cancel"])
        instance.item_count = instance.stockrequestitem_set.count()
        instance.save(update_fields=["item_count"])


@receiver(post_save, sender=StockRequestItem, dispatch_uid="stock_request_item_on_post_save")
def stock_request_item_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
) -> None:
    if not raw and not update_fields:
        instance.stock_request.item_count = StockRequestItem.objects.filter(
            stock_request=instance.stock_request
        ).count()
        instance.stock_request.save(update_fields=["item_count"])


@receiver(post_save, sender=RepackRequest, dispatch_uid="repack_request_on_post_save")
def repack_request_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
) -> None:
    if not raw and not update_fields:
        result = get_task_result(instance)
        if getattr(result, "state", "") == PENDING:
            pass
        else:
            task = run_task_sync_or_async(
                process_repack_request,
                repack_request_id=str(instance.id),
                username=instance.user_modified or instance.user_created,
            )
            instance.task_id = getattr(task, "id", None)
            instance.save(update_fields=["task_id"])


@receiver(
    post_save,
    sender=Allocation,
    dispatch_uid="allocation_on_post_save",
)
def allocation_on_post_save(sender, instance, raw, created, update_fields, **kwargs) -> None:
    stock = Stock.objects.get(code=instance.code)
    stock.subject_identifier = instance.registered_subject.subject_identifier
    stock.save(update_fields=["subject_identifier"])


@receiver(
    post_save,
    sender=StockTransferItem,
    dispatch_uid="stock_transfer_item_on_post_save",
)
def stock_transfer_item_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
) -> None:
    if not raw and not update_fields:
        instance.stock.in_transit = True
        instance.stock.save(update_fields=["in_transit"])


@receiver(
    post_save,
    sender=ConfirmationAtLocationItem,
    dispatch_uid="confirm_at_location_item_on_post_save",
)
def confirm_at_location_item_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
) -> None:
    if not raw and not update_fields:
        instance.stock.confirmed_at_location = True
        instance.stock.save(update_fields=["confirmed_at_location"])


@receiver(
    post_save,
    sender=StorageBinItem,
    dispatch_uid="storage_bin_item_on_post_save",
)
def storage_bin_item_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
) -> None:
    if not raw and not update_fields:
        instance.stock.stored_at_location = True
        instance.stock.save(update_fields=["stored_at_location"])


@receiver(
    post_save,
    sender=DispenseItem,
    dispatch_uid="dispense_item_on_post_save",
)
def dispense_item_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
) -> None:
    if not raw and not update_fields:
        instance.stock.dispensed = True
        instance.stock.qty_out = 1
        instance.stock.unit_qty_out = instance.stock.container_unit_qty * 1
        instance.stock.save(update_fields=["dispensed", "qty_out", "unit_qty_out"])
        StorageBinItem.objects.filter(stock=instance.stock).delete()


@receiver(post_delete, sender=Confirmation, dispatch_uid="confirmation_on_post_delete")
def confirmation_on_post_delete(ender, instance, using, **kwargs):
    """Update confirmed"""
    instance.stock.confirmed = False
    instance.stock.save(update_fields=["confirmed"])


@receiver(post_delete, sender=ReceiveItem, dispatch_uid="receive_item_on_post_delete")
def receive_item_on_post_delete(sender, instance, using, **kwargs) -> None:
    instance.order_item.unit_qty_received = (
        instance.order_item.unit_qty_received - instance.unit_qty_received
    )
    instance.order_item.save()


@receiver(post_delete, sender=Stock, dispatch_uid="stock_on_post_delete")
def stock_on_post_delete(sender, instance, using, **kwargs) -> None:
    if getattr(instance, "confirmation", None):
        instance.confirm.delete()


@receiver(
    pre_delete,
    sender=Allocation,
    dispatch_uid="allocation_pre_delete",
)
def allocation_pre_delete(sender, instance, using, **kwargs) -> None:
    if getattr(instance, "stock", None):
        try:
            obj = StockTransferItem.objects.get(code=instance.stock.code)
        except ObjectDoesNotExist:
            pass
        else:
            raise ProtectedError(
                _(
                    "%s object can't be deleted because it has already been transferred "
                    "to another location. See %s"
                )
                % (instance._meta.object_name, obj.stock_transfer.transfer_identifier),
                {instance},
            )


@receiver(
    post_delete,
    sender=Allocation,
    dispatch_uid="allocation_post_delete",
)
def allocation_post_delete(sender, instance, using, **kwargs) -> None:
    if stock := getattr(instance, "stock", None):
        stock.subject_identifier = NULL_STRING
        stock.allocation = None
        stock.save(update_fields=["subject_identifier", "allocation"])


@receiver(
    post_delete,
    sender=StockTransferItem,
    dispatch_uid="stock_transfer_item_post_delete",
)
def stock_transfer_item_post_delete(sender, instance, using, **kwargs) -> None:
    instance.stock.in_transit = False
    instance.stock.save(update_fields=["in_transit"])


@receiver(
    post_delete,
    sender=ConfirmationAtLocationItem,
    dispatch_uid="confirm_at_location_item_post_delete",
)
def confirm_at_location_item_post_delete(sender, instance, using, **kwargs) -> None:
    instance.stock.confirmed_at_location = False
    instance.stock.save(update_fields=["confirmed_at_location"])


@receiver(
    post_delete,
    sender=StorageBinItem,
    dispatch_uid="storage_bin_item_post_delete",
)
def storage_bin_item_post_delete(sender, instance, using, **kwargs) -> None:
    instance.stock.stored_at_location = False
    instance.stock.save(update_fields=["stored_at_location"])


@receiver(
    post_delete,
    sender=DispenseItem,
    dispatch_uid="dispense_item_on_post_delete",
)
def dispense_item_on_post_delete(sender, instance, using, **kwargs) -> None:
    instance.stock.dispensed = False
    instance.stock.qty_out = 0
    instance.stock.unit_qty_out = 0
    instance.stock.save(update_fields=["dispensed", "qty_out", "unit_qty_out"])


@receiver(
    post_save,
    dispatch_uid="create_or_update_refills_on_post_save",
)
def create_or_update_refills_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
):
    if not raw:
        try:
            instance.related_visit_model_attr()  # see edc-visit-tracking
        except AttributeError:
            pass
        else:
            try:
                instance.creates_refills_from_crf()
            except AttributeError as e:
                if "creates_refills_from_crf" not in str(e):
                    raise
                pass


@receiver(
    post_save,
    dispatch_uid="update_refill_end_datetime",
)
def update_previous_refill_end_datetime_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
):
    if not raw and not update_fields and isinstance(instance, (StudyMedicationCrfModelMixin,)):
        update_previous_refill_end_datetime(instance)



# Generated by Django 2.1 on 2018-08-03 03:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("sites", "0002_alter_domain_unique"), ("edc_auth", "0001_initial")]

    operations = [
        migrations.RemoveField(model_name="userprofile", name="country"),
        migrations.RemoveField(model_name="userprofile", name="study_country"),
        migrations.RemoveField(model_name="userprofile", name="study_site"),
        migrations.AddField(
            model_name="userprofile",
            name="sites",
            field=models.ManyToManyField(to="sites.Site"),
        ),
    ]



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "edc_offstudy"
    verbose_name = "Edc Offstudy"
    has_exportable_data = True
    include_in_administration_section = False



# Generated by Django 5.2.8 on 2025-11-26 03:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_dx_review",
            "0004_remove_diagnosislocations_edc_dx_revi_name_a39b40_idx_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="diagnosislocations",
            name="extra_value",
            field=models.CharField(default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="diagnosislocations",
            name="field_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                help_text="Not required",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="diagnosislocations",
            name="plural_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Plural name"
            ),
        ),
        migrations.AlterField(
            model_name="historicaldiagnosislocations",
            name="extra_value",
            field=models.CharField(default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="historicaldiagnosislocations",
            name="field_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                help_text="Not required",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="historicaldiagnosislocations",
            name="plural_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Plural name"
            ),
        ),
        migrations.AlterField(
            model_name="historicalreasonsfortesting",
            name="extra_value",
            field=models.CharField(default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="historicalreasonsfortesting",
            name="field_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                help_text="Not required",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="historicalreasonsfortesting",
            name="plural_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Plural name"
            ),
        ),
        migrations.AlterField(
            model_name="reasonsfortesting",
            name="extra_value",
            field=models.CharField(default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="reasonsfortesting",
            name="field_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                help_text="Not required",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="reasonsfortesting",
            name="plural_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Plural name"
            ),
        ),
    ]



from django.core.management.base import BaseCommand

from edc_auth.export_users import export_users


class Command(BaseCommand):
    help = "Export users to a CSV file in the current directory"

    def add_arguments(self, parser):
        parser.add_argument("--csvfile", default=None, dest="csvfile", help="CSV filename")

        parser.add_argument(
            "--verbose",
            default=False,
            action="store_true",
            dest="verbose",
            help="Verbose mode",
        )

        parser.add_argument(
            "--export",
            default=False,
            action="store_true",
            dest="export_to_file",
            help="Export new users to file",
        )

    def handle(self, *args, **options):  # noqa: ARG002
        export_users(options["csvfile"])



from django import forms

from ...models import Container


class ContainerForm(forms.ModelForm):
    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("max_items_per_subject") and not cleaned_data.get(
            "may_request_as"
        ):
            raise forms.ValidationError(
                {"max_items_per_subject": "Not applicable. Leave blank or 0"}
            )
        if not cleaned_data.get("max_items_per_subject") and cleaned_data.get(
            "may_request_as"
        ):
            raise forms.ValidationError({"max_items_per_subject": "This field is required"})
        if (
            cleaned_data.get("max_items_per_subject")
            and cleaned_data.get("max_items_per_subject") > 6
            and cleaned_data.get("may_request_as")
        ):
            raise forms.ValidationError({"max_items_per_subject": "May not exceed 6"})
        return cleaned_data

    class Meta:
        model = Container
        fields = "__all__"



# Generated by Django 2.2 on 2019-06-08 05:29

import uuid

import _socket
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
import simple_history.models
from django.conf import settings
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("edc_visit_schedule", "0002_auto_20190305_0123"),
    ]

    operations = [
        migrations.CreateModel(
            name="VisitSchedule",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("visit_schedule_name", models.CharField(max_length=150)),
                ("schedule_name", models.CharField(max_length=150)),
                ("visit_code", models.CharField(max_length=150)),
                ("visit_name", models.CharField(max_length=150)),
                ("visit_title", models.CharField(max_length=150)),
                ("timepoint", models.IntegerField()),
                ("active", models.BooleanField(default=False)),
            ],
            options={"ordering": ("visit_schedule_name", "schedule_name", "visit_code")},
        ),
        migrations.CreateModel(
            name="HistoricalVisitSchedule",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("visit_schedule_name", models.CharField(max_length=150)),
                ("schedule_name", models.CharField(max_length=150)),
                ("visit_code", models.CharField(max_length=150)),
                ("visit_name", models.CharField(max_length=150)),
                ("visit_title", models.CharField(max_length=150)),
                ("timepoint", models.IntegerField()),
                ("active", models.BooleanField(default=False)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical visit schedule",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



# Generated by Django 5.2.10 on 2026-01-24 15:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0109_rename_qty_decimal_places_container_qty_suggested_places_and_more",
        ),
    ]

    operations = [
        migrations.RenameField(
            model_name="container",
            old_name="qty_suggested",
            new_name="qty_default",
        ),
        migrations.RenameField(
            model_name="container",
            old_name="qty_suggested_places",
            new_name="qty_places",
        ),
        migrations.RenameField(
            model_name="historicalcontainer",
            old_name="qty_suggested",
            new_name="qty_default",
        ),
        migrations.RenameField(
            model_name="historicalcontainer",
            old_name="qty_suggested_places",
            new_name="qty_places",
        ),
    ]



from .home_view import AeHomeView
from .tmg import (
    ClosedTmgAeListboardView,
    DeathListboardView,
    NewTmgAeListboardView,
    OpenTmgAeListboardView,
    SummaryListboardView,
    TmgHomeView,
)



from __future__ import annotations

from django import forms

from ...models import OrderItem


class OrderItemForm(forms.ModelForm):
    def clean(self):
        if (
            self.cleaned_data.get("container")
            and self.cleaned_data.get("container").unit_qty_max is None
        ):
            raise forms.ValidationError(
                {
                    "container": (
                        "Invalid. Container maximum unit quantity has not been set. "
                        "Please update the container before continuing."
                    )
                }
            )
        if (
            self.cleaned_data.get("container")
            and self.cleaned_data.get("container_unit_qty") is not None
            and self.cleaned_data.get("container").unit_qty_max
            < self.cleaned_data.get("container_unit_qty")
        ):
            raise forms.ValidationError(
                {
                    "container_unit_qty": (
                        "Invalid. Container unit quantity may not exceed "
                        f"{self.cleaned_data.get('container').unit_qty_max}"
                    )
                }
            )

        if self.instance.id and self.instance.unit_qty_received > self.cleaned_data.get(
            "item_qty_ordered"
        ):
            raise forms.ValidationError(
                {
                    "item_qty_ordered": (
                        "Invalid. May not be less than the unit quantity already received "
                    )
                }
            )

    class Meta:
        model = OrderItem
        fields = "__all__"



# Generated by Django 2.0.1 on 2018-01-16 13:28

import django.db.models.deletion
from django.db import migrations, models

import edc_sites.models


class Migration(migrations.Migration):
    dependencies = [
        ("sites", "0002_alter_domain_unique"),
        ("edc_action_item", "0002_auto_20180109_1117"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="actionitem",
            managers=[("on_site", edc_sites.models.CurrentSiteManager())],
        ),
        migrations.AddField(
            model_name="actionitem",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="historicalactionitem",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
    ]



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "edc_view_utils"
    verbose_name = "Edc View Utils"
    has_exportable_data = False
    include_in_administration_section = False



from django.core.handlers.wsgi import WSGIRequest

from edc_randomization.blinding import user_is_blinded_from_request

from .list_filters import AssignmentListFilter


def remove_fields_for_blinded_users(request: WSGIRequest, fields: tuple) -> tuple[str, ...]:
    """You need to secure custom SimpleListFilters yourself"""
    if user_is_blinded_from_request(request):
        fields = list(fields)
        for f in fields:
            if isinstance(f, str):
                if "assignment" in f or "lot_no" in f or "lot" in f:
                    fields.remove(f)
            elif isinstance(f, tuple):
                f, _ = f
                if "assignment" in f or "lot_no" in f or "lot" in f:
                    fields.remove(f)
            elif issubclass(f, AssignmentListFilter):
                fields.remove(f)
        fields = tuple(fields)
    return fields



from __future__ import annotations

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import admin
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_identifier import SubjectIdentifierError, is_subject_identifier_or_raise

from ..actions import flag_as_verified_against_paper, unflag_as_verified_against_paper


class ConsentModelAdminMixin:
    name_fields: tuple[str] = ("first_name", "last_name")
    name_display_field: str = "first_name"
    actions = (flag_as_verified_against_paper, unflag_as_verified_against_paper)

    def __init__(self, *args):
        self.update_radio_fields()
        super().__init__(*args)

    def update_radio_fields(self) -> None:
        self.radio_fields.update(
            {
                "language": admin.VERTICAL,
                "gender": admin.VERTICAL,
                "is_dob_estimated": admin.VERTICAL,
                "identity_type": admin.VERTICAL,
                "is_incarcerated": admin.VERTICAL,
                "may_store_samples": admin.VERTICAL,
                "consent_reviewed": admin.VERTICAL,
                "study_questions": admin.VERTICAL,
                "assessment_score": admin.VERTICAL,
                "consent_copy": admin.VERTICAL,
                "is_literate": admin.VERTICAL,
            }
        )

    def get_fields(self, request, obj=None) -> tuple[str, ...]:
        original_fields = super().get_fields(request, obj=obj)
        return (
            "subject_identifier",
            *self.name_fields,
            "initials",
            "language",
            "is_literate",
            "witness_name",
            "consent_datetime",
            "gender",
            "dob",
            "is_dob_estimated",
            "guardian_name",
            "identity",
            "identity_type",
            "confirm_identity",
            "is_incarcerated",
            "may_store_samples",
            "comment",
            "consent_reviewed",
            "study_questions",
            "assessment_score",
            "consent_copy",
            *original_fields,
        )

    def get_readonly_fields(self, request, obj=None) -> tuple[str, ...]:
        readonly_fields = super().get_readonly_fields(request, obj)
        fields = ("subject_identifier", "subject_identifier_as_pk")
        if obj:
            return (
                *fields,
                "consent_datetime",
                "identity",
                "confirm_identity",
                *readonly_fields,
            )
        return fields + readonly_fields

    def get_search_fields(self, request) -> tuple[str, ...]:
        search_fields: tuple[str] = super().get_search_fields(request)
        return tuple(
            {*search_fields, "id", "subject_identifier", *self.name_fields, "identity"}
        )

    def get_list_display(self, request) -> tuple[str, ...]:
        list_display: tuple[str] = super().get_list_display(request)
        custom_fields = (
            "subject_identifier",
            "screening_identifier",
            "is_verified",
            "is_verified_datetime",
            self.name_display_field,
            "initials",
            "gender",
            "dob",
            "may_store_samples",
            "consent_datetime",
            "created",
            "modified",
            "user_created",
            "user_modified",
        )
        if request.user.has_perm("edc_data_manager.add_dataquery"):
            custom_fields = list(custom_fields)
            custom_fields.insert(3, self.queries)
            custom_fields = tuple(custom_fields)
        return *custom_fields, *tuple(f for f in list_display if f not in custom_fields)

    def get_list_filter(self, request) -> tuple[str, ...]:
        list_filter = super().get_list_filter(request)
        custom_fields = (
            "gender",
            "is_verified",
            "is_verified_datetime",
            "language",
            "may_store_samples",
            "is_literate",
            "consent_datetime",
            "created",
            "modified",
            "user_created",
            "user_modified",
            "hostname_created",
        )
        return custom_fields + tuple(f for f in list_filter if f not in custom_fields)

    def delete_view(self, request, object_id, extra_context=None):
        """Prevent deletion if SubjectVisit objects exist."""
        extra_context = extra_context or {}
        subject_consent_model_cls = django_apps.get_model(settings.SUBJECT_CONSENT_MODEL)
        related_visit_model_cls = django_apps.get_model(settings.SUBJECT_VISIT_MODEL)
        obj = subject_consent_model_cls.objects.get(id=object_id)
        try:
            protected = [
                related_visit_model_cls.objects.get(subject_identifier=obj.subject_identifier)
            ]
        except ObjectDoesNotExist:
            protected = None
        except MultipleObjectsReturned:
            protected = related_visit_model_cls.objects.filter(
                subject_identifier=obj.subject_identifier
            )
        extra_context.update({"protected": protected})
        return super().delete_view(request, object_id, extra_context)

    def get_next_options(self, request=None, **kwargs) -> dict:
        """Returns the key/value pairs from the "next" querystring
        as a dictionary.
        """
        subject_screening_model_cls = django_apps.get_model(settings.SUBJECT_SCREENING_MODEL)
        next_options = super().get_next_options(request=request, **kwargs)
        try:
            is_subject_identifier_or_raise(next_options["subject_identifier"])
        except SubjectIdentifierError:
            next_options["subject_identifier"] = subject_screening_model_cls.objects.get(
                subject_identifier_as_pk=next_options["subject_identifier"]
            ).subject_identifier
        except KeyError:
            pass
        return next_options

    @admin.display(description="Open queries")
    def queries(self, obj=None) -> str:
        new_url = reverse(
            "edc_data_manager_admin:edc_data_manager_dataquery_add",
        )
        if obj.is_verified:
            formatted_html = None
        else:
            links = []
            for query_obj in obj.open_data_queries:
                url = reverse(
                    "edc_data_manager_admin:edc_data_manager_dataquery_change",
                    args=(query_obj.id,),
                )
                links.append(
                    f'<A title="go to query" href="{url}">'
                    f"{query_obj.action_identifier[-9:]}</A>"
                )
            if links:
                formatted_html = format_html(
                    '<BR>{links}<BR><A title="New query" href="{new_url}">Add query</A>',
                    links=mark_safe("<BR>".join(links)),  # nosec B703 B308  # noqa: S308
                    new_url=mark_safe(new_url),  # nosec B703 B308  # noqa: S308
                )
            else:
                formatted_html = format_html(
                    '<A title="New query" href="{new_url}?"'
                    'subject_identifier={subject_identifier}">Add query</A>',
                    new_url=mark_safe(new_url),  # nosec B703 B308  # noqa: S308
                    subject_identifier=obj.subject_identifier,
                )
        return formatted_html


class ModelAdminConsentMixin(ConsentModelAdminMixin):
    pass



from django.contrib import admin
from django.utils import timezone
from django_audit_fields import ModelAdminAuditFieldsMixin
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_model_admin.mixins import (
    ModelAdminHideDeleteButtonOnCondition,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminRedirectAllToChangelistMixin,
    ModelAdminRedirectOnDeleteMixin,
    TemplatesModelAdminMixin,
)
from edc_sites.admin import SiteModelAdminMixin

from ..admin_site import edc_export_admin
from ..forms import DataRequestForm
from ..models import DataRequest, DataRequestHistory
from .data_request_history_admin import DataRequestHistoryInline


@admin.register(DataRequest, site=edc_export_admin)
class DataRequestAdmin(
    SiteModelAdminMixin,
    TemplatesModelAdminMixin,
    ModelAdminRedirectOnDeleteMixin,
    ModelAdminRevisionMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminAuditFieldsMixin,
    ModelAdminRedirectAllToChangelistMixin,
    ModelAdminHideDeleteButtonOnCondition,
    admin.ModelAdmin,
):
    show_cancel = True
    view_on_site = False
    show_history_label = False

    change_search_field_name = "id"

    actions = ("export_selected",)

    ordering = ("-created",)

    inlines = [DataRequestHistoryInline]

    form = DataRequestForm

    date_hierarchy = "created"

    fields = ("name", "models", "export_format", "decrypt")

    list_display = (
        "name",
        "description",
        "export_format",
        "decrypt",
        "user_created",
        "created",
    )

    list_filter = ("user_created", "created", "decrypt", "export_format")

    search_fields = ("id", "models", "description", "name")

    def export_selected(self, request, queryset):
        for obj in queryset:
            DataRequestHistory.objects.create(data_request=obj)
            rows_updated = queryset.update(exported_datetime=timezone.now())
            if rows_updated == 1:
                message_bit = "1 data request was"
            else:
                message_bit = "%s data requests were" % rows_updated
            self.message_user(request, "%s successfully exported." % message_bit)

    export_selected.short_description = "Export selected data requests"

    def hide_delete_button_on_condition(self, request, object_id) -> bool:
        return True



from __future__ import annotations

from decimal import Decimal
from uuid import UUID

from celery import shared_task
from django.apps import apps as django_apps
from django.db import transaction
from django.utils import timezone

from ..exceptions import InsufficientStockError, RepackError


@shared_task
def process_repack_request(repack_request_id: UUID | None = None, username: str | None = None):
    """Take from stock and fill container as new stock item."""
    repack_request_model_cls = django_apps.get_model("edc_pharmacy.repackrequest")
    stock_model_cls = django_apps.get_model("edc_pharmacy.stock")
    repack_request = repack_request_model_cls.objects.get(id=repack_request_id)
    repack_request.task_id = None
    repack_request.item_qty_processed = repack_request.item_qty_processed = (
        stock_model_cls.objects.filter(repack_request=repack_request).count()
    )
    repack_request.item_qty_repack = (
        repack_request.item_qty_processed
        if not repack_request.item_qty_repack
        else repack_request.item_qty_repack
    )
    item_qty_to_process = repack_request.item_qty_repack - repack_request.item_qty_processed
    if not getattr(repack_request.from_stock, "confirmation", None):
        raise RepackError("Source stock item not confirmed")
    stock_model_cls = repack_request.from_stock.__class__
    with transaction.atomic():
        for _ in range(0, int(item_qty_to_process)):
            try:
                stock_model_cls.objects.create(
                    receive_item=None,
                    qty_in=1,
                    qty_out=0,
                    qty=1,
                    unit_qty_in=repack_request.container_unit_qty,
                    unit_qty_out=Decimal("0.0"),
                    from_stock=repack_request.from_stock,
                    container=repack_request.container,
                    container_unit_qty=repack_request.container_unit_qty,
                    location=repack_request.from_stock.location,
                    repack_request=repack_request,
                    lot=repack_request.from_stock.lot,
                    user_created=username,
                    created=timezone.now(),
                )
            except InsufficientStockError:
                break
            else:
                repack_request.item_qty_processed += 1
                repack_request.unit_qty_processed += repack_request.container_unit_qty
                repack_request.from_stock.unit_qty_out += repack_request.container_unit_qty
        repack_request.from_stock.user_modified = username
        repack_request.from_stock.save(update_fields=["unit_qty_out", "user_modified"])
        repack_request.from_stock.refresh_from_db()
        repack_request.user_modified = username
        repack_request.modified = timezone.now()
        repack_request.save(
            update_fields=[
                "item_qty_repack",
                "item_qty_processed",
                "unit_qty_processed",
                "task_id",
                "user_modified",
                "modified",
            ]
        )
        repack_request.refresh_from_db()
        # repack_request.from_stock.unit_qty_out = (
        #     repack_request.from_stock.unit_qty_out + repack_request.unit_qty_processed
        # )
        #
        # repack_request.from_stock.save(update_fields=["unit_qty_out"])


__all__ = ["process_repack_request"]



from __future__ import annotations

import shutil
import sys
from pathlib import Path
from tempfile import mkdtemp
from typing import TYPE_CHECKING

import pandas as pd
from django.apps import apps as django_apps
from django.db import OperationalError
from django.utils import timezone
from tabulate import tabulate
from tqdm import tqdm

from edc_model_to_dataframe.model_to_dataframe import ModelToDataframe
from edc_sites.site import sites

from .constants import CSV, STATA_14, STATA_15

if TYPE_CHECKING:
    from datetime import datetime

    from django.contrib.auth.base_user import AbstractBaseUser
    from django.contrib.auth.models import AnonymousUser, User
    from pandas import pd

    from edc_data_manager.models import DataDictionary


class ModelsToFileError(Exception):
    pass


class ModelsToFileNothingExportedError(Exception):
    pass


class ModelsToFile:
    """Exports a list of models to individual CSV files and
    adds each to a single zip archive.

    models: a list of model names in label_lower format.
    """

    date_format: str = "%Y-%m-%d %H:%M:%S"
    delimiter: str = "|"
    encoding: str = "utf-8"

    def __init__(
        self,
        *,
        user: User | AbstractBaseUser | AnonymousUser,
        models: list[str],
        export_folder: Path | None = None,
        site_ids: list[int] | None = None,
        decrypt: bool | None = None,
        archive_to_single_file: bool | None = None,
        export_format: str | int | None = None,
        use_simple_filename: bool | None = None,
        date_format: str | None = None,
    ):
        self.archive_filename: str | None = None
        self.emailed_datetime: datetime | None = None
        self.emailed_to: str | None = None
        self.exported_filenames: list = []
        self.export_history = dict(model=[], rows=[])
        self.export_format = export_format or CSV
        if export_format not in [CSV, STATA_14, STATA_15]:
            raise ModelsToFileError(
                f"Invalid export format. Expected one of {[CSV, STATA_14, STATA_15]}. "
                f"Got {export_format}"
            )
        self.use_simple_filename = use_simple_filename
        self.date_format = date_format or self.date_format

        self.archive_to_single_file: bool = (
            True if archive_to_single_file is None else archive_to_single_file
        )
        self.decrypt: bool = decrypt or False
        self.models: list[str] = models or []
        self.user = user

        self.site_ids = site_ids or [sites.get_current_site().site_id]
        for site_id in self.site_ids:
            if not sites.get_site_ids_for_user(user=self.user, site_id=site_id):
                self.site_ids = [s for s in self.site_ids if s != site_id]

        if export_folder and not export_folder.exists():
            raise ModelsToFileError(f"Export folder does not exist. Got {export_folder}.")
        self.export_folder: Path = export_folder if export_folder else Path(mkdtemp())
        formatted_date: str = timezone.now().strftime("%Y%m%d%H%M%S")
        self.sub_folder = (
            f"{self.user.username}_{'csv' if export_format == CSV else 'stata'}_"
            f"{formatted_date}"
        )
        (self.export_folder / self.sub_folder).mkdir(parents=False, exist_ok=False)

        for model in tqdm(self.models, total=len(self.models)):
            if filename := self.model_to_file(model):
                self.exported_filenames.append(filename)
        if not self.exported_filenames:
            raise ModelsToFileNothingExportedError(f"Nothing exported. Got models={models}.")
        if self.export_history:
            with (self.export_folder / self.sub_folder / "README.txt").open("w") as f:
                f.write("\nExport Summary\n")
                f.write(
                    tabulate(
                        pd.DataFrame(data=self.export_history)
                        .sort_values("model")
                        .reset_index(drop=True),
                        tablefmt="fancy_grid",
                        headers="keys",
                        showindex=False,
                    )
                )
                f.write("\n")
                f.write(f"{timezone.now().strftime('%Y-%m-%d %H:%M:%S')}\n")
        if self.archive_to_single_file:
            self.archive_filename = self.create_archive_file()
            sys.stdout.write(f"{self.archive_filename}\n")

    def model_to_file(self, model: str) -> str | None:
        """Convert model to a dataframe and export as CSV or STATA
        using pandas.Dataframe to_csv() or to_stata().
        """
        filename = None
        try:
            dataframe = ModelToDataframe(
                model=model,
                decrypt=self.decrypt,
                sites=self.site_ids,
                drop_sys_columns=False,
                drop_action_item_columns=True,
                read_frame_verbose=False,
                remove_timezone=True,
            ).dataframe
        except OperationalError as e:
            if "1142" not in str(e):
                raise
            sys.stdout.write(f"Skipping. Got {e}\n")
        else:
            if not dataframe.empty:
                self.export_history["model"].append(model)
                self.export_history["rows"].append(len(dataframe))
                fname = (
                    model.split(".")[-1:][0].upper() if self.use_simple_filename else model
                ).replace(".", "_")
                if self.export_format == CSV:
                    path = self.export_folder / self.sub_folder / f"{fname}.csv"
                    dataframe.to_csv(
                        path_or_buf=path,
                        index=False,
                        encoding=self.encoding,
                        sep=self.delimiter,
                        date_format=self.date_format,
                    )
                elif self.export_format in [STATA_14, STATA_15]:
                    path = self.export_folder / self.sub_folder / f"{fname}.dta"
                    dataframe.to_stata(
                        path,
                        data_label=str(path),
                        version=118,
                        variable_labels=self.stata_variable_labels(dataframe, model=model),
                        write_index=False,
                    )
                else:
                    raise ModelsToFileNothingExportedError(
                        "Invalid file format. Expected CSV or STATA"
                    )
                filename = path.name
            return filename
        return None

    def get_filename_without_ext(self, model_name: str) -> str:
        return (
            model_name.split("_")[-1:][0].upper() if self.use_simple_filename else model_name
        )

    def create_archive_file(self):
        return shutil.make_archive(
            str(self.export_folder / self.sub_folder),
            "zip",
            root_dir=self.export_folder,
            base_dir=self.sub_folder,
        )

    def stata_variable_labels(self, dataframe: pd.DataFrame, model: str) -> dict[str, str]:
        variable_labels = dict(id="primary key")
        qs = self.data_dictionary_model_cls.objects.values("field_name", "prompt").filter(
            model=model, field_name__in=list(dataframe.columns)
        )
        variable_labels.update({obj.get("field_name"): obj.get("prompt")[:79] for obj in qs})
        return variable_labels

    @property
    def data_dictionary_model_cls(self) -> type[DataDictionary]:
        return django_apps.get_model("edc_data_manager.datadictionary")



from ..crf import OffstudyCrfModelFormMixin


class OffstudyRequisitionModelFormMixin(OffstudyCrfModelFormMixin):
    """ModelForm mixin for Requisition Models."""

    report_datetime_field_attr = "requisition_datetime"



from clinicedc_constants import OPEN, OTHER
from django.db import models
from django.db.models.deletion import PROTECT
from django.utils import timezone

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_search.model_mixins import SearchSlugManager, SearchSlugModelMixin
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin

from ..constants import PACKED, SHIPPED, STORAGE, TESTING, VERIFIED
from ..identifiers import BoxIdentifier
from ..model_mixins import VerifyBoxModelMixin
from .box_type import BoxType

BOX_DIMENSIONS = (("8 x 8", "8 x 8"), ("9 x 9", "9 x 9"), ("10 x 10", "10 x 10"))

BOX_CATEGORY = ((TESTING, "Testing"), (STORAGE, "Storage"), (OTHER, "Other"))

STATUS = (
    (OPEN, "Open"),
    (VERIFIED, "Verified"),
    (PACKED, "Packed"),
    (SHIPPED, "Shipped"),
)

human_readable_pattern = r"^[A-Z]{3}\-[0-9]{4}\-[0-9]{2}$"


class BoxIsFullError(Exception):
    pass


class BoxManager(SearchSlugManager, models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, box_identifier):
        return self.get(box_identifier=box_identifier)


class Box(SearchSlugModelMixin, VerifyBoxModelMixin, SiteModelMixin, BaseUuidModel):
    search_slug_fields = ("box_identifier", "human_readable_identifier", "name")

    box_identifier = models.CharField(max_length=25, editable=False, unique=True)

    name = models.CharField(max_length=25, null=True, blank=True)

    box_datetime = models.DateTimeField(default=timezone.now)

    box_type = models.ForeignKey(BoxType, on_delete=PROTECT)

    category = models.CharField(max_length=25, default=TESTING, choices=BOX_CATEGORY)

    category_other = models.CharField(max_length=25, null=True, blank=True)

    specimen_types = models.CharField(
        max_length=25,
        help_text=(
            "List of specimen types in this box. Use two-digit numeric "
            "codes separated by commas."
        ),
    )

    status = models.CharField(max_length=15, default=OPEN, choices=STATUS)

    accept_primary = models.BooleanField(
        default=False,
        help_text="Tick to allow 'primary' specimens to be added to this box",
    )

    comment = models.TextField(blank=True, default="")

    objects = BoxManager()

    on_site = CurrentSiteManager()

    history = HistoricalRecords()

    def save(self, *args, **kwargs):
        if not self.box_identifier:
            self.box_identifier = BoxIdentifier().identifier
        if not self.name:
            self.name = self.box_identifier
        if self.id:
            self.update_verified()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name

    def natural_key(self):
        return (self.box_identifier,)

    natural_key.dependencies = ["edc_lab.boxtype", "sites.Site"]

    @property
    def count(self):
        return self.boxitem_set.all().count()

    @property
    def items(self):
        return self.boxitem_set.all().order_by("position")

    @property
    def human_readable_identifier(self):
        x = self.box_identifier
        return f"{x[0:4]}-{x[4:8]}-{x[8:12]}"

    @property
    def next_position(self):
        """Returns an integer or None."""
        last_obj = self.boxitem_set.all().order_by("position").last()
        if not last_obj:
            next_position = 1
        else:
            next_position = last_obj.position + 1
        if next_position > self.box_type.total:
            raise BoxIsFullError(
                f"Box is full. Box {self.human_readable_identifier} has "
                f"{self.box_type.total} specimens."
            )
        return next_position

    @property
    def max_position(self):
        return

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Box"
        verbose_name_plural = "Boxes"



codenames = [
    "edc_refusal.add_subjectrefusal",
    "edc_refusal.change_subjectrefusal",
    "edc_refusal.delete_subjectrefusal",
    "edc_refusal.view_historicalsubjectrefusal",
    "edc_refusal.view_subjectrefusal",
]



from __future__ import annotations

import contextlib
from copy import copy
from typing import Any

from django.core.exceptions import NON_FIELD_ERRORS
from django.forms import ValidationError

APPLICABLE_ERROR = "applicable"
INVALID_ERROR = "invalid"
NOT_APPLICABLE_ERROR = "not_applicable"
NOT_REQUIRED_ERROR = "not_required"
REQUIRED_ERROR = "required"
OUT_OF_RANGE_ERROR = "out_of_range"


class InvalidModelFormFieldValidator(Exception):  # noqa: N818
    def __init__(self, message, code=None):
        message = f"Invalid field validator. Got '{message}'"
        super().__init__(message)
        self.code = code


class ModelFormFieldValidatorError(Exception):
    def __init__(self, message, code=None):
        super().__init__(message)
        self.code = code


class BaseFormValidator:
    default_fk_stored_field_name: str = "name"
    default_fk_display_field_name: str = "display_name"

    def __init__(
        self,
        cleaned_data: dict = None,
        instance: Any = None,
        data: dict = None,
        model: Any = None,  # model class
    ) -> None:
        self._errors: dict = {}
        self._error_codes: list = []
        self.cleaned_data = {}
        self.original_cleaned_data = {}
        self.data = {}
        if cleaned_data is None:
            raise ModelFormFieldValidatorError(
                f"{self!r}. Expected a cleaned_data dictionary. Got None."
            )
        self.cleaned_data = copy(cleaned_data)
        self.original_cleaned_data = copy(cleaned_data)
        if data:
            self.data = copy(data)
        self.instance = instance
        self.model = model
        try:
            self.instance.id
        except AttributeError:
            self.add_form = True
            self.change_form = False
        else:
            self.add_form = False
            self.change_form = True

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(cleaned_data={self.cleaned_data})"

    def __str__(self) -> str:
        return str(self.cleaned_data)

    def clean(self) -> None:
        """Override with logic normally in ModelForm.clean()"""
        pass

    def _clean(self) -> None:
        self.clean()

    def update_cleaned_data_from_instance(self, field: str) -> None:
        if field in self.original_cleaned_data:
            raise ModelFormFieldValidatorError(
                f"May not get form field value from instance. "
                f"This field is already in cleaned data. "
                f"Got {field}."
            )
        self.cleaned_data.update({field: getattr(self.instance, field)})

    def get(self, field: str, inline_set=None) -> Any:
        """Returns the field`s value"""
        if inline_set:
            field_value = self.get_inline_field_value(field, inline_set)
        else:
            try:
                field_value = self.cleaned_data.get(field).name
            except AttributeError:
                field_value = self.cleaned_data.get(field)
        return field_value

    def raise_validation_error(
        self, message: dict[str, str] | str, error_code: str, exc: Exception | None = None
    ) -> None:
        if isinstance(message, str):
            message = {NON_FIELD_ERRORS: message}
        self._errors.update(message)
        self._error_codes.append(error_code)
        if exc:
            raise ValidationError(message, code=error_code or INVALID_ERROR) from exc
        raise ValidationError(message, code=error_code or INVALID_ERROR)

    def validate(self) -> dict:
        """Call in ModelForm.clean.

        For example:

            form_validator_cls = None

            def clean(self):
                cleaned_data = super().clean()
                form_validator = self.form_validator_cls(
                    cleaned_data=cleaned_data)
                cleaned_data = form_validator.validate()
                return cleaned_data

        """
        try:
            self._clean()
        except ValidationError as e:
            self.capture_error_message(e)
            self.capture_error_code(e)
            raise ValidationError(e) from e
        return self.cleaned_data

    def capture_error_message(self, e: ValidationError) -> None:
        try:
            self._errors.update(**e.error_dict)
        except AttributeError:
            try:
                self._errors.update({NON_FIELD_ERRORS: e.error_list})
            except AttributeError:
                self._errors.update({NON_FIELD_ERRORS: str(e)})

    def capture_error_code(self, e: ValidationError) -> None:
        with contextlib.suppress(AttributeError):
            self._error_codes.append(e.code)

    def get_inline_field_value(self, field=None, inline_set=None):
        """Returns the value of the first inline field that has a value."""
        field_values = []
        total_forms = int(self.data.get(f"{inline_set}-TOTAL_FORMS"))
        for i in range(0, total_forms):
            if self.data.get(f"{inline_set}-{i}-DELETE") == "on":
                pass
            else:
                inline_field = f"{inline_set}-{i}-{field}"
                try:
                    field_values.append(self.data[inline_field])
                except KeyError:
                    break
        field_value = None
        for value in field_values:
            try:
                field_value = value.name
            except AttributeError:
                field_value = value
            if field_value:
                break
        return field_value



from __future__ import annotations

from typing import TYPE_CHECKING

from clinicedc_constants import INCOMPLETE, NOT_APPLICABLE
from django.core.exceptions import ObjectDoesNotExist

from edc_metadata import KEYED
from edc_metadata.metadata_helper import MetadataHelperMixin
from edc_visit_tracking.constants import MISSED_VISIT, SCHEDULED, UNSCHEDULED
from edc_visit_tracking.utils import (
    get_subject_visit_missed_model,
    get_subject_visit_missed_model_cls,
)

from .choices import APPT_TIMING
from .constants import MISSED_APPT, SCHEDULED_APPT, UNSCHEDULED_APPT
from .exceptions import (
    AppointmentBaselineError,
    AppointmentReasonUpdaterCrfsExistsError,
    AppointmentReasonUpdaterError,
    AppointmentReasonUpdaterRequisitionsExistsError,
    UnscheduledAppointmentError,
)
from .utils import get_allow_skipped_appt_using, raise_on_appt_may_not_be_missed

if TYPE_CHECKING:
    from edc_visit_tracking.model_mixins import VisitModelMixin

    from .models import Appointment


class AppointmentReasonUpdater(MetadataHelperMixin):
    """A class to try to update `reason` field based on the
    response from appointment timing and appointment reason.

    See also the Appointment signal that creates
    the related visit instance, if it does not exist
    when appt_timing==MISSED_APPT. (missed_appointment)

    See also the Appointment signal that deletes
    a related visit for a cancelled appointment.
    (cancelled_appointment)

    """

    metadata_helper_instance_attr: str = "appointment"

    def __init__(
        self,
        *,
        appointment: Appointment,
        appt_timing: str,
        appt_reason: str,
        commit: bool | None = None,
    ):
        self._related_visit = None
        self.appointment = appointment
        if not getattr(self.appointment, "id", None):
            raise AppointmentReasonUpdaterError(
                "Appointment instance must exist. Got `id` is None"
            )
        self.commit = commit
        self.appt_timing = appt_timing or self.appointment.appt_timing
        if self.appt_timing not in [a for a, b in self.get_appt_timing_choices()]:
            raise AppointmentReasonUpdaterError(
                f"Invalid value for appt_timing. "
                f"Expected on of {[a for a, b in self.get_appt_timing_choices()]}. "
                f"Got {self.appt_timing}"
            )
        try:
            raise_on_appt_may_not_be_missed(
                appointment=appointment, appt_timing=self.appt_timing
            )
        except (AppointmentBaselineError, UnscheduledAppointmentError) as e:
            raise AppointmentReasonUpdaterError(e) from e

        self.appt_reason = appt_reason or self.appointment.appt_reason

        self.update_related_visit_or_raise()

    def update_related_visit_or_raise(self) -> None:
        if self.related_visit:
            if self.appt_timing == MISSED_APPT and self.appt_reason in [
                SCHEDULED_APPT,
                UNSCHEDULED_APPT,
            ]:
                self.update_related_visit_to_missed_or_raise()
            elif self.appt_timing != MISSED_APPT and self.appt_reason in [
                SCHEDULED_APPT,
                UNSCHEDULED_APPT,
            ]:
                self.update_related_visit_to_not_missed_or_raise()
            else:
                raise AppointmentReasonUpdaterError(
                    f"Condition not handled. Got appt_reason={self.appt_reason}, "
                    f"appt_timing={self.appt_timing}"
                )

    @property
    def related_visit(self) -> VisitModelMixin | None:
        if not self._related_visit:
            try:
                self._related_visit = getattr(
                    self.appointment, self.appointment.related_visit_model_attr()
                )
            except ObjectDoesNotExist:
                self._related_visit = None
            except AttributeError as e:
                if "related_visit_model_attr" not in str(e):
                    raise
                self._related_visit = None
        return self._related_visit

    def update_related_visit_to_missed_or_raise(self) -> None:
        self.raise_if_keyed_data_exists()
        if self.related_visit:
            if self.appt_timing != MISSED_APPT:
                raise AppointmentReasonUpdaterError(
                    f"Appointment is not missed. Got appt_timing=`{self.appt_timing}` "
                    f"See {self.appointment}."
                )
            self.related_visit.reason = MISSED_VISIT
            self.related_visit.document_status = INCOMPLETE
            if self.commit:
                # the signal to update metadata is called on post_save
                self.related_visit.save_base(
                    update_fields=["reason", "document_status", "comments"]
                )
                self.related_visit.refresh_from_db()

    def update_related_visit_to_not_missed_or_raise(self) -> None:
        """Updates the subject visit instance from MISSED_VISIT
        to SCHEDULED or UNSCHEDULED.
        """
        if self.related_visit:
            reason = self.get_reason_from_appt_reason(self.appt_reason)
            if self.appt_timing == MISSED_APPT:
                raise AppointmentReasonUpdaterError(
                    f"Appointment is missed. Got appt_timing=`{self.appt_timing}` "
                    f"See {self.appointment}."
                )
            self.delete_subject_visit_missed_if_exists()
            self.related_visit.reason = reason
            self.related_visit.document_status = INCOMPLETE
            if self.related_visit.comments:
                self.related_visit.comments = self.related_visit.comments.replace(
                    "[auto-created]", ""
                )
            if self.commit:
                # the signal to update metadata is called on post_save
                # see also `update_document_status_on_save`
                self.related_visit.save_base(
                    update_fields=["reason", "document_status", "comments"]
                )
                self.related_visit.refresh_from_db()

    def raise_if_keyed_data_exists(self) -> None:
        """Raises an exception if CRF or Requisition metadata exists"""
        if self.crf_metadata_keyed_exists:
            if not self.crf_metadata_keyed_is_subject_visit_missed_only:
                raise AppointmentReasonUpdaterCrfsExistsError(
                    "Invalid. CRFs have already been entered for this timepoint."
                )
        elif self.requisition_metadata_keyed_exists:
            raise AppointmentReasonUpdaterRequisitionsExistsError(
                "Invalid. Requisitions have already been entered for this timepoint."
            )

    @property
    def crf_metadata_keyed_is_subject_visit_missed_only(self) -> bool:
        """Returns True if the only keyed CRF is the subject
        visit missed model.
        """
        return (
            self.get_crf_metadata_by(KEYED).count() == 1
            and self.get_crf_metadata_by(KEYED)[0].model == get_subject_visit_missed_model()
        )

    def delete_subject_visit_missed_if_exists(self) -> None:
        """Deletes the subject visit missed report if it exists"""
        get_subject_visit_missed_model_cls().objects.filter(
            subject_visit__appointment=self.appointment
        ).delete()

    def get_reason_from_appt_reason(self, appt_reason: str) -> str:
        """Returns a subject visit reason given the appt reason"""
        if appt_reason == SCHEDULED_APPT:
            visit_reason = SCHEDULED
        elif appt_reason == UNSCHEDULED_APPT and self.appointment.visit_code_sequence > 0:
            visit_reason = UNSCHEDULED
        else:
            raise AppointmentReasonUpdaterError(
                "Update failed. This is not an unscheduled appointment. "
                f"Got visit_code_sequence > 0 for appt_reason={appt_reason}"
            )
        return visit_reason

    @staticmethod
    def get_appt_timing_choices():
        if not get_allow_skipped_appt_using():
            return tuple([tpl for tpl in APPT_TIMING if tpl[0] != NOT_APPLICABLE])
        return APPT_TIMING
